# from ContainerHandle.dataProcess.config import _union_index

import warnings
from datetime import datetime as dtm
from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np
from pandas import DataFrame, concat, DatetimeIndex
# from scipy.interpolate import interp1d
from scipy.interpolate import UnivariateSpline as unvpline, interp1d

warnings.filterwarnings("ignore")

__all__ = ['merge_SMPS_APS']


def _powerlaw_fit(_coeA, _coeB, _aps, _idx, _factor):
    # breakpoint()

    _smps_fit_df = _coeA * (_aps.keys().values / _factor) ** _coeB
    return DataFrame(((_smps_fit_df.copy() - _aps.copy()) ** 2).sum(axis=1), columns=[_idx])


## Calculate S2
## 1. SMPS and APS power law fitting
## 2. shift factor from 0.5 ~ 3
## 3. calculate S2
## return : S2
# def _S2_calculate_dN(_smps, _aps):
def _powerlaw_fit_dN(_smps, _aps, _alg_type):
    print(f"\t\t\t{dtm.now().strftime('%m/%d %X')} : \033[92moverlap range fitting : {_alg_type}\033[0m")

    ## overlap fitting
    ## parmeter
    _dt_indx = _smps.index

    ## use SMPS data apply power law fitting
    ## y = Ax^B, A = e**coefa, B = coefb, x = logx, y = logy
    ## ref : http://mathworld.wolfram.com/LeastSquaresFittingPowerLaw.html
    ## power law fit to SMPS num conc at upper bins to log curve

    ## coefficient A, B
    _smps_qc_cond = ((_smps != 0) & np.isfinite(_smps))
    _smps_qc = _smps.where(_smps_qc_cond)

    _size = _smps_qc_cond.sum(axis=1)
    _size = _size.where(_size != 0.).copy()

    _logx, _logy = np.log(_smps_qc.keys()._data.astype(float)), np.log(_smps_qc)
    _x, _y, _xy, _xx = _logx.sum(), _logy.sum(axis=1), (_logx * _logy).sum(axis=1), (_logx ** 2).sum()

    _coeB = ((_size * _xy - _x * _y) / (_size * _xx - _x ** 2.))
    _coeA = np.exp((_y - _coeB * _x) / _size).values.reshape(-1, 1)
    _coeB = _coeB.values.reshape(-1, 1)

    ## rebuild shift smps data by coe. A, B
    ## x_shift = (y_ori/A)**(1/B)
    _aps_shift_x = (_aps / _coeA) ** (1 / _coeB)
    _aps_shift_x = _aps_shift_x.where(np.isfinite(_aps_shift_x))

    ## the least squares of diameter
    ## the shift factor which the closest to 1
    _shift_val = np.arange(0.3, 3.05, .05) ** .5
    # _shift_val = np.arange(0.9, 1.805, .005)**.5

    _shift_factor = DataFrame(columns=range(_shift_val.size), index=_aps_shift_x.index)
    _shift_factor.loc[:, :] = _shift_val

    # _dropna_idx = _shift_factor.dropna(how='all').index.copy()
    _dropna_idx = _aps_shift_x.dropna(how='all').index.copy()

    ## use the target function to get the similar aps and smps bin
    ## S2 = sum( (smps_fit_line(dia) - aps(dia*shift_factor) )**2 )
    ## assumption : the same diameter between smps and aps should get the same conc.

    ## be sure they art in log value
    _S2 = DataFrame(index=_aps_shift_x.index)
    _dia_table = DataFrame(np.full(_aps_shift_x.shape, _aps_shift_x.keys()),
                           columns=_aps_shift_x.keys(), index=_aps_shift_x.index)

    pool = Pool(cpu_count())

    _S2 = pool.starmap(partial(_powerlaw_fit, _coeA, _coeB, _aps), list(enumerate(_shift_val)))

    pool.close()
    pool.join()

    S2 = concat(_S2, axis=1)[np.arange(_shift_val.size)]
    # S2 /= S2.max(axis=1).to_frame().values

    shift_factor_dN = DataFrame(
        _shift_factor.loc[_dropna_idx].values[range(len(_dropna_idx)), S2.loc[_dropna_idx].idxmin(axis=1).values],
        index=_dropna_idx).reindex(_dt_indx).astype(float)

    shift_factor_dN = shift_factor_dN.mask((shift_factor_dN ** 2 < 0.6) | (shift_factor_dN ** 2 > 2.6))

    return shift_factor_dN


def _corr_fc(_aps_dia, _smps_dia, _smps_dn, _aps_dn, _smooth, _idx, _sh):
    ds_fc = lambda _dt: _dt * _dt.index ** 2 * np.pi
    dv_fc = lambda _dt: _dt * _dt.index ** 3 * np.pi / 6

    _aps_sh = _aps_dia / _sh
    _aps_sh_inp = _aps_sh.where((_aps_sh >= 500) & (_aps_sh <= 1500.)).copy()
    _aps_sh_corr = _aps_sh.where((_aps_sh >= _smps_dia[-1]) & (_aps_sh <= 1500.)).copy()

    corr_x = np.append(_smps_dia, _aps_sh_corr.dropna())

    input_x = np.append(_smps_dia, _aps_sh_inp.dropna())
    input_y = concat([_smps_dn, _aps_dn.iloc[:, ~np.isnan(_aps_sh_inp)]], axis=1)
    input_y.columns = input_x

    input_x.sort()
    input_y = input_y[input_x]
    corr_y = input_y[corr_x]

    S2_lst = []
    for (_tm, _inp_y_dn), (_tm, _cor_y_dn) in zip(input_y.dropna(how='all').iterrows(),
                                                  corr_y.dropna(how='all').iterrows()):
        ## corr(spec_data, spec_spline)
        _spl_dt = [unvpline(input_x, _inp_y, s=_smooth)(corr_x) for _inp_y in
                   [_inp_y_dn, ds_fc(_inp_y_dn), dv_fc(_inp_y_dn)]]
        _cor_dt = [_cor_y_dn, ds_fc(_cor_y_dn), dv_fc(_cor_y_dn)]

        _cor_all = sum([np.corrcoef(_cor, _spl)[0, 1] for _cor, _spl in zip(_cor_dt, _spl_dt)])

        S2_lst.append((3 - _cor_all) / 3)

    return DataFrame(S2_lst, columns=[_idx])


# def _S2_calculate_dSdV(_smps, _aps, _shft_dn, _S2, smps_ori, aps_ori):
# def _S2_calculate_dSdV(_smps, _aps, smps_ori=None):
def _corr_with_dNdSdV(_smps, _aps, _alg_type):
    print(f"\t\t\t{dtm.now().strftime('%m/%d %X')} : \033[92moverlap range correlation : {_alg_type}\033[0m")

    _smps_dia = _smps.keys().astype(float)
    _aps_dia = _aps.keys().astype(float)

    all_index = _smps.index.copy()
    qc_index = DatetimeIndex(set(_smps.dropna(how='all').index) & set(_aps.dropna(how='all').index)).sort_values()

    _smps_dn = _smps.loc[qc_index].copy()
    _aps_dn = _aps.loc[qc_index].copy()

    ds_fc = lambda _dt: _dt * _dt.index ** 2 * np.pi
    dv_fc = lambda _dt: _dt * _dt.index ** 3 * np.pi / 6

    _std_bin = np.geomspace(11.8, 19810, 230)
    _merge_bin = _std_bin[(_std_bin >= _smps_dia[-1]) & (_std_bin < 1500)].copy()

    _smooth = 50

    _shift_val = np.arange(0.9, 2.65, .05) ** .5

    ## spline fitting with shift aps and smps
    pool = Pool(cpu_count())

    S2_lst = pool.starmap(partial(_corr_fc, _aps_dia, _smps_dia, _smps_dn, _aps_dn, _smooth),
                          list(enumerate(_shift_val)))

    pool.close()
    pool.join()

    S2_table = concat(S2_lst, axis=1).set_index(qc_index)[np.arange(_shift_val.size)].astype(float).dropna()
    min_shft = S2_table.idxmin(axis=1).values

    return DataFrame(_shift_val[min_shft.astype(int)], index=S2_table.index).astype(float).reindex(_smps.index)


## Create merge data
##  shift all smps bin and remove the aps bin which smaller than the latest old smps bin
## Return : merge bins, merge data, density
def _merge_data(_smps_ori, _aps_ori, _shift_ori, _smps_lb, _aps_hb, _shift_mode, _alg_type):
    print(f"\t\t\t{dtm.now().strftime('%m/%d %X')} : \033[92mcreate merge data : {_shift_mode} and {_alg_type}\033[0m")

    _ori_idx = _smps_ori.index.copy()
    # _merge_idx = _smps_ori.loc[_aps_ori.dropna(how='all').index].dropna(how='all').index

    _corr_aps_cond = _aps_ori.keys() < 700
    _corr_aps_ky = _aps_ori.keys()[_corr_aps_cond]

    _merge_idx = DatetimeIndex(set(_smps_ori.dropna(how='all').index) & set(_aps_ori.dropna(how='all').index) &
                               set(_shift_ori.dropna(how='all').index)).sort_values()

    _smps, _aps, _shift = _smps_ori.loc[_merge_idx], _aps_ori.loc[_merge_idx], _shift_ori.loc[_merge_idx].values

    ## parameter
    _smps_key, _aps_key = _smps.keys()._data.astype(float), _aps.keys()._data.astype(float)

    _cntr = 1000
    _bin_lb = _smps_key[-1]

    ## make shift bins
    _smps_bin = np.full(_smps.shape, _smps_key)
    _aps_bin = np.full(_aps.shape, _aps_key)

    _std_bin = np.geomspace(_smps_key[0], _aps_key[-1], 230)
    _std_bin_merge = _std_bin[(_std_bin < _cntr) & (_std_bin > _bin_lb)]
    _std_bin_inte1 = _std_bin[_std_bin <= _bin_lb]
    _std_bin_inte2 = _std_bin[_std_bin >= _cntr]

    if _shift_mode == 'mobility':
        _aps_bin /= _shift

    elif _shift_mode == 'aerodynamic':
        _smps_bin *= _shift

    ## merge
    _merge_lst, _corr_lst = [], []
    for _bin_smps, _bin_aps, _dt_smps, _dt_aps, _sh in zip(_smps_bin, _aps_bin, _smps.values, _aps.values, _shift):
        ## keep complete smps bins and data
        ## remove the aps bin data lower than smps bin
        _condi = _bin_aps >= _bin_smps[-1]

        _merge_bin = np.hstack((_bin_smps, _bin_aps[_condi]))
        _merge_dt = np.hstack((_dt_smps, _dt_aps[_condi]))

        _merge_fit_loc = (_merge_bin < 1500) & (_merge_bin > _smps_lb)

        ## coeA and coeB
        _unvpl_fc = unvpline(np.log(_merge_bin[_merge_fit_loc]), np.log(_merge_dt[_merge_fit_loc]), s=50)
        _inte_fc = interp1d(_merge_bin, _merge_dt, kind='linear', fill_value='extrapolate')

        _merge_dt_fit = np.hstack((_inte_fc(_std_bin_inte1), np.exp(_unvpl_fc(np.log(_std_bin_merge))),
                                   _inte_fc(_std_bin_inte2)))

        _merge_lst.append(_merge_dt_fit)
        _corr_lst.append(interp1d(_std_bin, _merge_dt_fit)(_bin_aps[_corr_aps_cond]))

    _df_merge = DataFrame(_merge_lst, columns=_std_bin, index=_merge_idx)
    _df_merge = _df_merge.mask(_df_merge < 0)

    _df_corr = DataFrame(_corr_lst, columns=_corr_aps_ky, index=_merge_idx) / _aps_ori.loc[_merge_idx, _corr_aps_ky]

    ## process output df
    ## average, align with index
    def _out_df(*_df_arg, **_df_kwarg):
        _df = DataFrame(*_df_arg, **_df_kwarg).reindex(_ori_idx)
        _df.index.name = 'time'
        return _df

    return _out_df(_df_merge), _out_df(_shift_ori ** 2), _out_df(_df_corr)


def _fitness_func(psd, rho, pm25):
    psd_pm25 = psd[psd.keys()[psd.keys().values <= 2500]] * np.diff(np.log10(psd.keys())).mean()
    rho_pm25 = pm25 / (psd_pm25 * np.pi * psd_pm25.keys().values ** 3 / 6 * 1e-9).sum(axis=1, min_count=1)

    return (rho['density'] - rho_pm25) ** 2


def merge_SMPS_APS(df_smps, df_aps, df_pm25, aps_unit='um', smps_overlap_lowbound=500, aps_fit_highbound=1000,
                   dndsdv_alg=True, times_range=(0.8, 1.25, .05)):
    # merge_data, merge_data_dn, merge_data_dsdv, merge_data_cor_dn, density, density_dn, density_dsdv, density_cor_dn = [DataFrame([np.nan])] * 8

    ## set to the same units
    smps, aps = df_smps.copy(), df_aps.copy()
    smps.columns = smps.keys().to_numpy(float)
    aps.columns = aps.keys().to_numpy(float)

    if aps_unit == 'um':
        aps.columns = aps.keys() * 1e3

    fitness_typ = dict(dn=[], cor_dn=[], dndsdv=[], cor_dndsdv=[])
    shift_typ = dict(dn=[], cor_dn=[], dndsdv=[], cor_dndsdv=[])
    oth_typ = dict()

    times_ary = np.arange(*times_range).round(4)
    # times_ary = np.arange(*(0.8, 0.9, .05)).round(4)

    for times in times_ary:

        print(f"\t\t{dtm.now().strftime('%m/%d %X')} : \033[92mSMPS times value : {times}\033[0m")

        aps_input = aps.copy()
        aps_over = aps_input.loc[:, (aps.keys() > 700) & (aps.keys() < 1000)].copy()

        smps_input = (smps * times).copy()
        smps_over = smps_input[smps.keys()[smps.keys() > 500]].copy()

        for _count in range(2):

            ## shift data calculate
            ## original
            if _count == 0:
                alg_type = 'dn'
                shift = _powerlaw_fit_dN(smps_over, aps_over, alg_type)

                if dndsdv_alg:
                    shift_dsdv = _corr_with_dNdSdV(smps_over, aps_over, 'dndsdv').mask(shift.isna())

            ## aps correct
            else:
                alg_type = 'cor_dndsdv'
                shift_cor = _powerlaw_fit_dN(smps_over, aps_over, 'cor_dn')

                if dndsdv_alg:
                    shift = _corr_with_dNdSdV(smps_over, aps_over, alg_type).mask(shift_cor.isna())

            ## merge aps and smps
            ## 1. power law fit (dn) -> return dn data and aps correct factor
            ## 2. correaltion with dn, ds, dv -> return corrected dn_ds_dv data
            if (alg_type == 'dn') | dndsdv_alg:
                merge_arg = (smps_input, aps_input, shift, smps_overlap_lowbound, aps_fit_highbound)

                merge_data, density, _corr = _merge_data(*merge_arg, 'mobility', _alg_type=alg_type)
                density.columns = ['density']

                fitness_typ[alg_type].append(_fitness_func(merge_data, density, df_pm25))
                shift_typ[alg_type].append(shift[0])

            ## without aps correct
            if _count == 0:
                ## merge aps and smps
                ## dn_ds_dv data
                if dndsdv_alg:
                    alg_type = 'dndsdv'
                    merge_arg = (smps_input, aps_input, shift_dsdv, smps_overlap_lowbound, aps_fit_highbound)

                    merge_data_dsdv, density_dsdv, _ = _merge_data(*merge_arg, 'mobility', _alg_type=alg_type)
                    density_dsdv.columns = ['density']

                    fitness_typ[alg_type].append(_fitness_func(merge_data_dsdv, density_dsdv, df_pm25))
                    shift_typ[alg_type].append(shift_dsdv[0])

                ## dn data
                merge_data_dn, density_dn = merge_data.copy(), density.copy()

                ## correct aps data
                corr = _corr.resample('1d').mean().reindex(smps.index).ffill()
                corr = corr.mask(corr < 1, 1)

                aps_input.loc[:, corr.keys()] *= corr
                aps_over = aps_input.copy()


            ## with aps correct
            else:
                ## merge aps and smps
                ## dn data
                alg_type = 'cor_dn'
                merge_arg = (smps_input, aps_input, shift_cor, smps_overlap_lowbound, aps_fit_highbound)

                merge_data_cor_dn, density_cor_dn, _ = _merge_data(*merge_arg, 'mobility', _alg_type=alg_type)
                density_cor_dn.columns = ['density']

                fitness_typ[alg_type].append(_fitness_func(merge_data_cor_dn, density_cor_dn, df_pm25))
                shift_typ[alg_type].append(shift_cor[0])

    ## get times value and shift value
    out_dic = {}
    for (_typ, _lst), (_typ, _shft) in zip(fitness_typ.items(), shift_typ.items()):
        oth_typ[_typ] = None
        if len(_lst) == 0: continue

        df_times_min = concat(_lst, axis=1, keys=range(len(_lst))).idxmin(axis=1).dropna().astype(int)
        df_shift = concat(_shft, axis=1, keys=times_ary.tolist()).loc[df_times_min.index].values[
            range(len(df_times_min.index)), df_times_min.values]

        oth_typ[_typ] = DataFrame(np.array([df_shift, times_ary[df_times_min.values]]).T,
                                  index=df_times_min.index, columns=['shift', 'times']).reindex(smps.index)

    ## re-calculate merge_data
    alg_type = ['dn', 'cor_dn', 'dndsdv', 'cor_dndsdv'] if dndsdv_alg else ['dn', 'cor_dn']

    out_dic = {}
    den_lst, times_lst = [], []
    for _typ in alg_type:
        print(f"\t\t{dtm.now().strftime('%m/%d %X')} : \033[92mre-caculate merge data with times: {_typ}\033[0m")
        typ = oth_typ[_typ]
        smps_input = smps.copy() * typ['times'].to_frame().values

        corr_typ = corr if 'cor' in _typ else 1
        aps_input = aps.copy()
        aps_input.loc[:, corr.keys()] *= corr_typ

        merge_arg = (smps_input, aps_input, typ['shift'].to_frame(), smps_overlap_lowbound, aps_fit_highbound)

        merge_data, density, _corr = _merge_data(*merge_arg, 'mobility', _alg_type=_typ)
        density.columns = ['density']

        out_dic[f'data_{_typ}'] = merge_data

        den_lst.append(density)
        times_lst.append(typ['times'])

    out_rho = concat(den_lst, axis=1)
    out_times = concat(times_lst, axis=1)
    out_rho.columns = alg_type
    out_times.columns = alg_type

    # breakpoint()

    ## out
    out_dic.update(dict(density=out_rho, times=out_times))

    # out_dic = {
    # 'data_cor_dndsdv' : merge_data,
    # 'data_dn'     : merge_data_dn,
    # 'data_dndsdv' : merge_data_dsdv,
    # 'data_cor_dn' : merge_data_cor_dn,

    # 'density' : out_rho,

    # 'data_all_aer' : merge_data_aer,

    # 'density_cor_dndsdv' : density,
    # 'density_dn'   		 : density_dn,
    # 'density_dndsdv'	 : density_dsdv,
    # 'density_cor_dn'	 : density_cor_dn,
    # }

    ## process data
    for _nam, _df in out_dic.items():
        out_dic[_nam] = _df.reindex(smps.index).copy()

    return out_dic



"""
Pytest configuration and fixtures for AeroViz tests.
"""
import shutil
import tempfile
from datetime import datetime
from pathlib import Path

import pytest

# =============================================================================
# Path Configuration
# =============================================================================

# Base path for test fixtures
FIXTURES_PATH = Path(__file__).parent / 'fixtures'
RAW_DATA_PATH = FIXTURES_PATH / 'raw_data'

# Supported instruments with reader tests (EPA/Minion handle custom file
# formats, not physical instruments — they intentionally have no test_readers
# coverage)
INSTRUMENTS = [
    'AE33', 'AE43', 'APS', 'Aurora', 'BAM1020', 'BC1054',
    'GRIMM', 'IGAC', 'MA350', 'NEPH',
    'OCEC', 'Q-ACSM', 'SMPS', 'TEOM', 'VOC', 'Xact'
]


# =============================================================================
# Fixtures
# =============================================================================

@pytest.fixture(scope='session')
def fixtures_path():
    """Return the path to test fixtures directory."""
    return FIXTURES_PATH


@pytest.fixture(scope='session')
def raw_data_path():
    """Return the path to raw data fixtures directory."""
    return RAW_DATA_PATH


@pytest.fixture(scope='function')
def temp_output_dir():
    """Create a temporary directory for test outputs, cleaned up after test."""
    temp_dir = tempfile.mkdtemp(prefix='aeroviz_test_')
    yield Path(temp_dir)
    # Cleanup after test
    shutil.rmtree(temp_dir, ignore_errors=True)


@pytest.fixture(scope='session')
def default_date_range():
    """Default date range for testing (covers most test data)."""
    return {
        'start': datetime(2024, 1, 1),
        'end': datetime(2024, 12, 31, 23, 59, 59)
    }


# =============================================================================
# Instrument-specific fixtures
# =============================================================================

def _get_instrument_data_path(instrument: str) -> Path:
    """Get the data path for a specific instrument."""
    return RAW_DATA_PATH / instrument


def _has_test_data(instrument: str) -> bool:
    """Check if an instrument has test data files (including in subdirectories)."""
    path = _get_instrument_data_path(instrument)
    if not path.exists():
        return False
    # Check for any data files recursively (including in scenario subdirectories)
    return any(f.is_file() for f in path.rglob('*'))


# Generate fixtures for each instrument
for _instrument in INSTRUMENTS:
    # Create a fixture for each instrument's data path
    exec(f'''
@pytest.fixture(scope='session')
def {_instrument.lower().replace('-', '_')}_data_path():
    """Return the path to {_instrument} test data."""
    path = _get_instrument_data_path('{_instrument}')
    if not _has_test_data('{_instrument}'):
        pytest.skip('{_instrument} test data not available')
    return path
''')


# =============================================================================
# Markers
# =============================================================================

def pytest_configure(config):
    """Register custom markers."""
    config.addinivalue_line(
        "markers", "requires_data: mark test as requiring actual data files"
    )
    config.addinivalue_line(
        "markers", "slow: mark test as slow running"
    )
    config.addinivalue_line(
        "markers", "reader: rawDataReader pipeline test (slow, I/O heavy)"
    )
    config.addinivalue_line(
        "markers", "dataprocess: post-processing test under AeroViz.dataProcess.*"
    )
    config.addinivalue_line(
        "markers",
        "optical: optical-property test (Mie, IMPROVE, BrC, RI retrieval)",
    )
    # Add markers for each instrument (normalize Q-ACSM -> q_acsm to match test files)
    for instrument in INSTRUMENTS:
        marker_name = instrument.lower().replace('-', '_')
        config.addinivalue_line(
            "markers", f"{marker_name}: mark test as {instrument} specific"
        )


def pytest_collection_modifyitems(config, items):
    """Auto-apply ``reader`` + ``slow`` markers to every test under
    ``tests/test_readers/``. Saves the per-file boilerplate of repeating
    ``pytestmark = pytest.mark.reader`` in 16 instrument modules.
    """
    reader_marker = pytest.mark.reader
    slow_marker = pytest.mark.slow
    for item in items:
        path = str(item.fspath)
        if "tests/test_readers/" in path or "tests\\test_readers\\" in path:
            item.add_marker(reader_marker)
            item.add_marker(slow_marker)


# =============================================================================
# Helpers available to all tests
# =============================================================================

@pytest.fixture(scope='session')
def get_instrument_files():
    """Return a function to get all test files for an instrument."""
    def _get_files(instrument: str, pattern: str = '*') -> list[Path]:
        path = _get_instrument_data_path(instrument)
        if not path.exists():
            return []
        return sorted(path.glob(pattern))
    return _get_files


@pytest.fixture(scope='session')
def get_instrument_scenarios():
    """
    Return a function to get test scenarios for an instrument.

    Scenarios are defined by subdirectories in the instrument's data folder:
    - normal/: Standard data files
    - multi_header/: Files with multiple headers (APS, etc.)
    - transposed/: Transposed format files
    - edge_cases/: Edge case files
    """
    def _get_scenarios(instrument: str) -> dict[str, Path]:
        path = _get_instrument_data_path(instrument)
        if not path.exists():
            return {}

        scenarios = {}
        for subdir in path.iterdir():
            if subdir.is_dir():
                scenarios[subdir.name] = subdir

        # If no subdirectories, treat root as 'normal' scenario
        if not scenarios:
            root_files = [f for f in path.iterdir() if f.is_file()]
            if root_files:
                scenarios['normal'] = path

        return scenarios
    return _get_scenarios



"""Smoke tests for the AeroViz.size top-level functions."""
import numpy as np
import pandas as pd
import pytest

from AeroViz.size import psd_stats, psd_distributions, merge_psd


pytestmark = pytest.mark.dataprocess


@pytest.fixture
def smps_like():
    """Simulated SMPS-like dN/dlogDp, 24 hours × 50 bins (12-500 nm)."""
    dp = np.logspace(np.log10(12), np.log10(500), 50)
    n_times = 24
    # Lognormal peak ~120 nm
    base = 1e5 * np.exp(-0.5 * (np.log(dp / 120) / np.log(1.8)) ** 2)
    return pd.DataFrame(
        np.tile(base, (n_times, 1)), columns=dp,
        index=pd.date_range('2024-01-01', periods=n_times, freq='h'),
    )


class TestPsdStats:
    def test_returns_weighting_keys(self, smps_like):
        out = psd_stats(smps_like, bin_range=(12, 500))
        assert {'number', 'surface', 'volume', 'other'}.issubset(out)

    def test_other_has_properties(self, smps_like):
        out = psd_stats(smps_like, bin_range=(12, 500))
        # 'other' wraps statistics — should be a DataFrame
        assert isinstance(out['other'], pd.DataFrame)


class TestPsdDistributions:
    def test_three_distributions(self, smps_like):
        out = psd_distributions(smps_like)
        for key in ('number', 'surface', 'volume', 'properties'):
            assert key in out
        # Surface/volume must be larger-diameter-weighted versions of number
        n_tot = out['number'].sum(axis=1).mean()
        v_tot = out['volume'].sum(axis=1).mean()
        assert v_tot > 0 and n_tot > 0


class TestMergePsd:
    def test_v4_requires_pm25(self, smps_like):
        with pytest.raises(ValueError, match="df_pm25"):
            merge_psd(smps_like, smps_like, version=4)

    def test_invalid_version(self, smps_like):
        with pytest.raises(ValueError, match="version must be one of"):
            merge_psd(smps_like, smps_like, version=99)



"""Top-level optical-processing functions; convenience wrappers — see
`AeroViz.dataProcess.Optical.*` for full algorithm details."""

from __future__ import annotations

import numpy as np
import pandas as pd

__all__ = [
    # Bulk extinction / Mie
    'optical_basic',
    'improve',
    'gas_extinction',
    'retrieve_ri',
    'brown_carbon',
    'mie',
    # One-shot lognormal entries
    'mie_lognormal',
    'mie_multimodal',
    # Angular scattering
    'scattering_function',
    'scattering_function_sd',
    'phase_matrix',
    'nephelometer_truncation_correction',
    # Core-shell Mie (Aden-Kerker)
    'mie_core_shell',
    'mie_core_shell_sd',
    # Inverse problems (RI retrieval)
    'iterative_inversion',
    'iterative_inversion_sd',
    'contour_intersection',
]


def optical_basic(df_sca, df_abs, df_mass=None, df_no2=None, df_temp=None):
    """Compute basic optical properties (extinction, SSA, MEE/MSE/MAE, Ångström
    exponents) from measured scattering and absorption.

    Parameters
    ----------
    df_sca : DataFrame
        Scattering coefficient (Mm⁻¹).
    df_abs : DataFrame
        Absorption coefficient (Mm⁻¹).
    df_mass : DataFrame, optional
        PM mass concentration (μg/m³), used for mass efficiencies.
    df_no2 : DataFrame, optional
        NO2 concentration (ppb), used to subtract gas absorption.
    df_temp : DataFrame, optional
        Ambient temperature (°C), used in gas-extinction correction.

    Returns
    -------
    DataFrame
        Derived optical properties.
    """
    from AeroViz.dataProcess.Optical._extinction import _basic
    return _basic(df_sca, df_abs, df_mass, df_no2, df_temp)


def improve(df_mass, df_RH=None, method='revised', df_nh4_status=None,
            df_ext=None, oa_oc_ratio=1.8, upper_bounds=None):
    """Calculate extinction using the IMPROVE equation.

    Parameters
    ----------
    df_mass : DataFrame
        Reconstructed mass concentrations. Required columns depend on `method`:
        - 'revised' / 'modified': AS, AN, OM, Soil, SS, EC
        - 'localized':            AS, AN, POC, SOC, Soil, SS, EC
    df_RH : DataFrame, optional
        Relative humidity (%).
    method : {'revised', 'modified', 'localized'}, default 'revised'
        IMPROVE variant.
    df_nh4_status : DataFrame, optional
        NH4 status from chemistry reconstruction; rows flagged
        'Deficiency' are excluded.
    df_ext : DataFrame, optional
        Measured extinction with 'Scattering' and 'Absorption' columns
        (Mm⁻¹). Required for method='localized'.
    oa_oc_ratio : float, default 1.8
        OA/OC conversion ratio (method='localized' only).
    upper_bounds : dict, optional
        Upper bounds for MLR coefficients (method='localized' only).

    Returns
    -------
    dict
        Keys include 'dry', 'wet', 'ALWC', 'fRH'; method='localized'
        adds 'coefficients' and 'regression'.
    """
    from AeroViz.dataProcess.Optical._IMPROVE import revised, modified, localized

    if method == 'revised':
        return revised(df_mass, df_RH, df_nh4_status)
    if method == 'modified':
        return modified(df_mass, df_RH, df_nh4_status)
    if method == 'localized':
        if df_ext is None:
            raise ValueError(
                "method='localized' requires df_ext with 'Scattering' and "
                "'Absorption' columns to fit POA/SOA mass scattering "
                "efficiencies via MLR."
            )
        return localized(df_mass, df_ext, df_RH, df_nh4_status,
                         oa_oc_ratio=oa_oc_ratio, upper_bounds=upper_bounds)
    raise ValueError(
        f"method must be 'revised', 'modified', or 'localized', got '{method}'"
    )


def gas_extinction(df_no2, df_temp):
    """Calculate gas contribution to extinction (Rayleigh + NO2 absorption).

    Parameters
    ----------
    df_no2 : DataFrame
        NO2 concentration (ppb).
    df_temp : DataFrame
        Ambient temperature (°C).

    Returns
    -------
    DataFrame
        Columns: ScatteringByGas, AbsorptionByGas, ExtinctionByGas (Mm⁻¹).
    """
    from AeroViz.dataProcess.Optical._IMPROVE import gas_extinction as _gas_ext
    return _gas_ext(df_no2, df_temp)


def retrieve_ri(df_optical, df_pnsd, dlogdp=0.014, wavelength=550):
    """Retrieve complex refractive index from co-located optical and PSD data.

    Parameters
    ----------
    df_optical : DataFrame
        Optical data with Extinction, Scattering, and Absorption columns.
    df_pnsd : DataFrame
        Particle number size distribution.
    dlogdp : float, default 0.014
        Logarithmic bin width.
    wavelength : float, default 550
        Wavelength in nm.

    Returns
    -------
    DataFrame
        Retrieved RI with `re_real` and `re_imaginary` columns.
    """
    from AeroViz.dataProcess.Optical._retrieve_RI import retrieve_RI
    return retrieve_RI(df_optical, df_pnsd, dlogdp, wavelength)


def brown_carbon(df_abs, wavelengths=None, ref_wavelength=880, aae_bc=1.0):
    """Separate BC and BrC absorption using the AAE approach.

    Parameters
    ----------
    df_abs : DataFrame
        Absorption coefficients at multiple wavelengths (Mm⁻¹), with
        columns named like 'abs_370', 'abs_470', ..., 'abs_880'.
    wavelengths : list[int], optional
        Wavelengths to compute BrC for. Default: [370, 470, 520, 590, 660].
    ref_wavelength : int, default 880
        Reference wavelength assumed to be pure-BC absorption.
    aae_bc : float, default 1.0
        AAE assumed for Black Carbon.

    Returns
    -------
    DataFrame
        BC/BrC absorption, BrC fraction at each wavelength, and AAE_BrC.
    """
    from AeroViz.dataProcess.Optical._derived import calculate_BrC_absorption
    return calculate_BrC_absorption(
        df_abs=df_abs,
        wavelengths=wavelengths,
        ref_wavelength=ref_wavelength,
        aae_bc=aae_bc,
    )


# =============================================================================
# Mie consolidation
# =============================================================================

def _is_mixing_table(ri: pd.DataFrame) -> bool:
    """Return True iff `ri` is a DataFrame with at least one
    `*_volume_ratio` column."""
    return isinstance(ri, pd.DataFrame) and any(
        str(c).endswith('_volume_ratio') for c in ri.columns
    )


def _mixing_table_to_complex_ri(ri_table: pd.DataFrame,
                                wavelength: float) -> np.ndarray:
    """Compute an effective complex RI per row by volume-weighted average
    of the species refractive indices for the given mixing table."""
    from AeroViz.dataProcess.core import REFRACTIVE_INDEX

    # Pick the wavelength key whose value is numerically closest to the
    # requested wavelength so that 450/550 nm constants are reused naturally.
    available = {int(k): k for k in REFRACTIVE_INDEX}
    key = available[min(available, key=lambda w: abs(w - wavelength))]
    species_ri = REFRACTIVE_INDEX[key]

    eff = np.zeros(len(ri_table), dtype=complex)
    for col in ri_table.columns:
        col = str(col)
        if not col.endswith('_volume_ratio'):
            continue
        species = col[: -len('_volume_ratio')]
        if species not in species_ri:
            continue
        eff = eff + species_ri[species] * ri_table[col].values
    return eff


def mie(df_psd, ri, wavelength=550, mixing=None, distribution=False):
    """Compute Mie optical properties from a particle size distribution.

    Parameters
    ----------
    df_psd : DataFrame
        PSD with rows = time, columns = diameters (nm).
    ri : Series of complex | DataFrame
        Refractive index. EITHER:
          - Series of complex numbers, one per row → single-material Mie
          - DataFrame with columns ending in '_volume_ratio' → species mixing
            table (one column per species, values in [0, 1])
    wavelength : float, default 550
        Wavelength in nm.
    mixing : {'internal', 'external', 'both'}, optional
        Only used when ri is a mixing table.
        - 'internal' (default if not given): all species mixed within
          each particle
        - 'external': each species contributes independently
        - 'both': compute and return both
    distribution : bool, default False
        False → return total ext/sca/abs (Mm⁻¹) per row.
        True  → return per-bin dExt/dSca/dAbs (Mm⁻¹) distributions.

    Returns
    -------
    DataFrame | dict
        - mixing in (None/'internal'/'external'): DataFrame
        - mixing == 'both': dict {'internal': DataFrame, 'external': DataFrame}
    """
    from AeroViz.dataProcess.Optical.mie import (
        Mie_SD,
        internal_mixing,
        external_mixing,
        calculate_extinction_distribution,
    )

    if mixing is not None and mixing not in ('internal', 'external', 'both'):
        raise ValueError(
            f"mixing must be one of None, 'internal', 'external', 'both'; "
            f"got '{mixing}'"
        )

    # --- Detect single-material vs mixing-table ---
    if _is_mixing_table(ri):
        # Mixing-table dispatch
        if distribution:
            raise NotImplementedError(
                "distribution=True with a mixing-table RI is not yet "
                "supported. Reduce the table to a single complex RI per "
                "row (e.g. volume-weighted) and pass it as a Series."
            )

        chosen = mixing or 'internal'

        if chosen == 'internal':
            eff_ri = _mixing_table_to_complex_ri(ri, wavelength)
            return internal_mixing(
                df_psd,
                pd.Series(eff_ri, index=ri.index),
                wavelength=wavelength,
            )

        # external_mixing expects volume-fraction columns WITHOUT the
        # '_volume_ratio' suffix — strip it for the call.
        stripped = ri.rename(
            columns={
                c: str(c)[: -len('_volume_ratio')]
                for c in ri.columns
                if str(c).endswith('_volume_ratio')
            }
        )

        if chosen == 'external':
            return external_mixing(df_psd, stripped, wavelength=wavelength)

        # chosen == 'both'
        eff_ri = _mixing_table_to_complex_ri(ri, wavelength)
        internal_df = internal_mixing(
            df_psd, pd.Series(eff_ri, index=ri.index), wavelength=wavelength,
        )
        external_df = external_mixing(df_psd, stripped, wavelength=wavelength)
        return {'internal': internal_df, 'external': external_df}

    # --- Single-material path ---
    if mixing is not None:
        raise ValueError(
            "mixing= is only valid when ri is a mixing table (DataFrame "
            "with '*_volume_ratio' columns)."
        )

    # Accept Series-of-complex, or 1-column complex DataFrame, or numpy array
    if isinstance(ri, pd.DataFrame):
        if ri.shape[1] != 1 or not np.issubdtype(
            ri.iloc[:, 0].dtype, np.complexfloating
        ):
            raise ValueError(
                "Single-material RI must be a Series of complex numbers "
                "(or a single-column DataFrame of complex dtype). For a "
                "species mixing table, use columns ending in "
                "'_volume_ratio'."
            )
        ri_array = ri.iloc[:, 0].values
    elif isinstance(ri, pd.Series):
        ri_array = ri.values
    else:
        ri_array = np.asarray(ri)

    if distribution:
        # calculate_extinction_distribution returns dict with 'ext','sca',
        # 'abs','diameter' as numpy arrays. Wrap into a tidy DataFrame
        # keyed by diameter and named by quantity, matching the
        # "per-bin dExt/dSca/dAbs distributions" contract.
        diameter = df_psd.columns.values.astype(float)
        result = calculate_extinction_distribution(
            refractive_index=ri_array,
            wavelength=wavelength,
            diameter=diameter,
            number_conc=df_psd.values,
        )

        # 1D PSD (single row) → squeeze yields (n_bins,)
        # 2D PSD → (n_times, n_bins). For consistency, always return a
        # dict of DataFrames keyed by quantity with rows=time, cols=Dp.
        ext_arr = np.atleast_2d(result['ext'])
        sca_arr = np.atleast_2d(result['sca'])
        abs_arr = np.atleast_2d(result['abs'])

        index = df_psd.index
        cols = diameter

        return {
            'ext': pd.DataFrame(ext_arr, index=index, columns=cols),
            'sca': pd.DataFrame(sca_arr, index=index, columns=cols),
            'abs': pd.DataFrame(abs_arr, index=index, columns=cols),
        }

    # Scalar ext/sca/abs per row via Mie_SD
    return Mie_SD(ri_array, wavelength, df_psd)


# =============================================================================
# Lognormal PSD shortcuts (Phase 2)
# =============================================================================

def mie_lognormal(refractive_index, wavelength=550, *, geo_mean=200,
                  geo_std=2.0, total_number=1e6, n_bins=167,
                  dp_range=(1, 2500)):
    """One-shot Mie optics from a single lognormal PSD.

    Convenience wrapper that builds the PSD internally then calls
    :func:`Mie_SD`. See :mod:`AeroViz.dataProcess.Optical.mie` for
    parameter details.

    Returns
    -------
    dict
        Keys ``ext``, ``sca``, ``abs`` (Mm⁻¹).
    """
    from AeroViz.dataProcess.Optical.mie import mie_lognormal as _impl
    return _impl(
        refractive_index, wavelength,
        geo_mean=geo_mean, geo_std=geo_std,
        total_number=total_number, n_bins=n_bins, dp_range=dp_range,
    )


def mie_multimodal(refractive_index, wavelength=550, *, modes,
                   n_bins=167, dp_range=(1, 2500)):
    """Mie optics for a multi-modal lognormal PSD.

    ``modes`` is a sequence of ``(geo_mean, geo_std, total_number)``
    tuples; their lognormal distributions are summed onto a shared
    diameter grid before the Mie integration.

    Returns
    -------
    dict
        Keys ``ext``, ``sca``, ``abs`` (Mm⁻¹).
    """
    from AeroViz.dataProcess.Optical.mie import mie_multimodal as _impl
    return _impl(
        refractive_index, wavelength,
        modes=modes, n_bins=n_bins, dp_range=dp_range,
    )


# =============================================================================
# Angular scattering (Phase 3)
# =============================================================================

def scattering_function(m, wavelength, diameter, angles=None, space='theta'):
    """Single-particle angular scattering pattern (Mie phase function).

    Returns dict with ``angles``, ``SL`` (parallel-polarised),
    ``SR`` (perpendicular), ``SU`` (unpolarised average).
    """
    from AeroViz.dataProcess.Optical.mie_angular import scattering_function as _impl
    return _impl(m, wavelength, diameter, angles=angles, space=space)


def scattering_function_sd(m, wavelength, dp, ndp, angles=None,
                           space='theta', psd_type='auto'):
    """PSD-integrated angular scattering pattern.

    Returns dict with the same keys as :func:`scattering_function`
    but PSD-integrated.
    """
    from AeroViz.dataProcess.Optical.mie_angular import scattering_function_sd as _impl
    return _impl(m, wavelength, dp, ndp, angles=angles,
                 space=space, psd_type=psd_type)


def phase_matrix(m, wavelength, diameter, mu=None):
    """Single-particle Mueller phase matrix (S₁₁, S₁₂, S₃₃, S₃₄).

    Returns dict with keys ``mu``, ``S11``, ``S12``, ``S33``, ``S34``.
    """
    from AeroViz.dataProcess.Optical.mie_angular import phase_matrix as _impl
    return _impl(m, wavelength, diameter, mu=mu)


def nephelometer_truncation_correction(sae, wavelength=550, instrument='NEPH'):
    """Anderson & Ogren (1998) truncation correction for nephelometers.

    Multiplicative factor to apply to uncorrected scattering for an
    integrating nephelometer (TSI 3563 default; Aurora 3000 also
    tabulated). ``sae`` is the scattering Ångström exponent.
    """
    from AeroViz.dataProcess.Optical.mie_angular import nephelometer_truncation_correction as _impl
    return _impl(sae, wavelength=wavelength, instrument=instrument)


# =============================================================================
# Core-shell Mie / Aden-Kerker theory (Phase 4)
# =============================================================================

def mie_core_shell(m_core, m_shell, d_core, d_total, wavelength):
    """Aden-Kerker coated-sphere Mie efficiencies (single particle).

    Returns dict with the same 7 keys as
    :func:`calculate_mie_efficiencies`: ``Q_ext``, ``Q_sca``, ``Q_abs``,
    ``g``, ``Q_pr``, ``Q_back``, ``Q_ratio``.
    """
    from AeroViz.dataProcess.Optical.mie_core_shell import mie_core_shell as _impl
    return _impl(m_core, m_shell, d_core, d_total, wavelength)


def mie_core_shell_sd(m_core, m_shell, dp_core, dp_total, ndp,
                     wavelength=550, psd_type='dNdlogDp'):
    """PSD-integrated Aden-Kerker coated-sphere Mie.

    Returns dict ``ext``, ``sca``, ``abs`` (Mm⁻¹) plus ``g_eff``
    (scattering-weighted mean asymmetry parameter).
    """
    from AeroViz.dataProcess.Optical.mie_core_shell import mie_core_shell_sd as _impl
    return _impl(m_core, m_shell, dp_core, dp_total, ndp,
                 wavelength=wavelength, psd_type=psd_type)


# =============================================================================
# Inverse problems / RI retrieval (Phase 5)
# =============================================================================

def iterative_inversion(b_ext, b_sca, b_abs, lognormal_params,
                        wavelength=550, n_initial=1.5, k_initial=0.01):
    """Newton-Raphson RI retrieval from measured (Bext, Bsca, Babs).

    PSD is described by ``lognormal_params``, a dict with keys
    ``geo_mean``, ``geo_std``, ``total_number``.

    Returns dict with ``n``, ``k``, ``iterations``, ``converged``,
    ``residuals``.
    """
    from AeroViz.dataProcess.Optical.mie_inverse import iterative_inversion as _impl
    return _impl(b_ext, b_sca, b_abs, lognormal_params,
                 wavelength=wavelength,
                 n_initial=n_initial, k_initial=k_initial)


def iterative_inversion_sd(b_ext, b_sca, b_abs, dp, ndp,
                            wavelength=550, n_initial=1.5, k_initial=0.01):
    """Newton-Raphson RI retrieval with explicit PSD (dp, ndp arrays)."""
    from AeroViz.dataProcess.Optical.mie_inverse import iterative_inversion_sd as _impl
    return _impl(b_ext, b_sca, b_abs, dp, ndp,
                 wavelength=wavelength,
                 n_initial=n_initial, k_initial=k_initial)


def contour_intersection(b_ext, b_sca, b_abs, lognormal_params,
                          wavelength=550, n_range=(1.3, 2.0),
                          k_range=(0, 0.5), grid=51):
    """Contour-intersection RI retrieval (Sumlin 2018 method)."""
    from AeroViz.dataProcess.Optical.mie_inverse import contour_intersection as _impl
    return _impl(b_ext, b_sca, b_abs, lognormal_params,
                 wavelength=wavelength,
                 n_range=n_range, k_range=k_range, grid=grid)



# Auto-import all instrument reader modules dynamically
# New instruments only need to:
# 1. Add a new .py file in this directory
# 2. Add entry to supported_instruments.py meta dict

import importlib
from pathlib import Path

# Get all .py files in this directory (excluding __init__.py)
_script_dir = Path(__file__).parent
_module_files = [
    f.stem for f in _script_dir.glob('*.py')
    if f.stem != '__init__' and not f.stem.startswith('_')
]

# Dynamically import each module
__all__ = []
for _module_name in _module_files:
    try:
        _module = importlib.import_module(f'.{_module_name}', package=__name__)
        globals()[_module_name] = _module
        __all__.append(_module_name)
    except ImportError as e:
        # Skip modules that fail to import (e.g., missing dependencies)
        pass

# Clean up temporary variables
del _script_dir, _module_files, _module_name, _module, importlib, Path



"""
Chemical calculations for aerosol analysis.

This module provides functions for:
- Molar concentration conversion
- Volume-average mixing refractive index
- Kappa (hygroscopicity parameter)
- Growth factor (gRH)
- Gas-particle partitioning ratios (SOR, NOR, NTR, epsilon)
"""

from pandas import concat, DataFrame

from AeroViz.dataProcess.core import REFRACTIVE_INDEX, safe_divide, validate_inputs

# =============================================================================
# Constants
# =============================================================================

# Molecular weights in g/mol
MOLECULAR_WEIGHTS = {
    'SO42-': 96.06,
    'NO3-': 62.00,
    'Cl-': 35.4,
    'Ca2+': 40.078,
    'K+': 39.098,
    'Mg2+': 24.305,
    'Na+': 22.99,
    'NH4+': 18.04,
}

# Gas molecular weights for partition calculations
GAS_MOLECULAR_WEIGHTS = {
    'SO2': 64.07,
    'NO2': 46.01,
    'NH3': 17.03,
    'HNO3': 63.01,
    'HCl': 36.46,
}

# =============================================================================
# Required columns definitions
# =============================================================================

VOLUME_MIXING_REQUIRED = ['total_dry']
VOLUME_MIXING_SPECIES = ['AS', 'AN', 'OM', 'Soil', 'SS', 'EC']

KAPPA_REQUIRED = ['gRH', 'AT', 'RH']
GRH_VOLUME_REQUIRED = ['total_dry']
GRH_ALWC_REQUIRED = ['ALWC']

# Partition calculation required columns
PARTITION_REQUIRED = ['temp']  # Temperature is required for molar conversion
PARTITION_SPECIES = {
    'SOR': ['SO42-', 'SO2'],      # Sulfur Oxidation Ratio
    'NOR': ['NO3-', 'NO2'],       # Nitrogen Oxidation Ratio
    'NOR_2': ['NO3-', 'NO2', 'HNO3'],  # NOR including HNO3
    'NTR': ['NH4+', 'NH3'],       # Nitrogen Transformation Ratio
    'epls_NO3': ['NO3-', 'HNO3'],  # NO3 partitioning
    'epls_NH4': ['NH4+', 'NH3'],   # NH4 partitioning
    'epls_Cl': ['Cl-', 'HCl'],     # Cl partitioning
}

# =============================================================================
# Column descriptions (Chinese/English) - 欄位說明
# =============================================================================

# Volume-average mixing 體積平均混合計算所需欄位
VOLUME_COLUMN_DESCRIPTIONS = {
    'AS_volume': 'Ammonium Sulfate volume (μm³/m³) 硫酸銨體積濃度',
    'AN_volume': 'Ammonium Nitrate volume (μm³/m³) 硝酸銨體積濃度',
    'OM_volume': 'Organic Matter volume (μm³/m³) 有機物體積濃度',
    'Soil_volume': 'Soil/Crustal volume (μm³/m³) 土壤/地殼物質體積濃度',
    'SS_volume': 'Sea Salt volume (μm³/m³) 海鹽體積濃度',
    'EC_volume': 'Elemental Carbon volume (μm³/m³) 元素碳體積濃度',
    'total_dry': 'Total dry aerosol volume (μm³/m³) 乾氣膠總體積濃度',
}

# Kappa 吸濕參數計算所需欄位
KAPPA_COLUMN_DESCRIPTIONS = {
    'gRH': 'Hygroscopic growth factor (Dp_wet/Dp_dry) 吸濕成長因子 (濕粒徑/乾粒徑)',
    'AT': 'Ambient Temperature (°C) 環境溫度',
    'RH': 'Relative Humidity (%) 相對濕度',
}

# gRH 成長因子計算所需欄位
GRH_COLUMN_DESCRIPTIONS = {
    'total_dry': 'Total dry aerosol volume (μm³/m³) 乾氣膠總體積濃度',
    'ALWC': 'Aerosol Liquid Water Content (μg/m³) 氣膠液態水含量',
}

# Partition 氣固分配比計算所需欄位
PARTITION_COLUMN_DESCRIPTIONS = {
    'temp': 'Ambient Temperature (°C) 環境溫度',
    'SO42-': 'Particulate Sulfate (μg/m³) 顆粒態硫酸鹽',
    'SO2': 'Gaseous Sulfur Dioxide (μg/m³) 氣態二氧化硫',
    'NO3-': 'Particulate Nitrate (μg/m³) 顆粒態硝酸鹽',
    'NO2': 'Gaseous Nitrogen Dioxide (μg/m³) 氣態二氧化氮',
    'HNO3': 'Gaseous Nitric Acid (μg/m³) 氣態硝酸',
    'NH4+': 'Particulate Ammonium (μg/m³) 顆粒態銨鹽',
    'NH3': 'Gaseous Ammonia (μg/m³) 氣態氨',
    'Cl-': 'Particulate Chloride (μg/m³) 顆粒態氯鹽',
    'HCl': 'Gaseous Hydrochloric Acid (μg/m³) 氣態鹽酸',
}


def convert_mass_to_molar_concentration(df):
    """
    Convert mass concentration (μg/m³) to molar concentration (μmol/m³ for particles, ppm for gases).

    This function identifies ionic species based on the MOLECULAR_WEIGHTS dictionary
    and converts them from mass to molar units. Gaseous species are converted using
    the ideal gas law with temperature data from the input DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing concentration data with column names matching ions
        in MOLECULAR_WEIGHTS. Must include 'temp' column in Celsius.

    Returns
    -------
    pandas.DataFrame
        DataFrame with all concentrations converted to molar units:
        - Ionic species: μg/m³ → μmol/m³
        - Gaseous species: μg/m³ → ppm (using ideal gas law)

    Notes
    -----
    - The function assumes temperature ('temp') is in Celsius and converts it to Kelvin
    - Uses the ideal gas constant of 0.082 L·atm/(mol·K)
    - Non-matched columns (except 'temp' and 'RH') are treated as gaseous species

    Examples
    --------
    >>> import pandas as pd
    >>> data = pd.DataFrame({
    ...     'SO42-': [10.0, 15.0],
    ...     'NO3-': [5.0, 7.5],
    ...     'O3': [30.0, 45.0],
    ...     'temp': [25.0, 30.0],
    ...     'RH': [60.0, 70.0]
    ... })
    >>> convert_mass_to_molar_concentration(data)
    """
    # Identify which columns are particulate ions vs. gases
    particle_keys = list(set(df.keys()) & set(MOLECULAR_WEIGHTS.keys()))
    gas_keys = list(set(df.keys()) - set(MOLECULAR_WEIGHTS.keys()) - {'temp', 'RH'})

    # Calculate gas constant * temperature factor for gas conversion (ideal gas law)
    temperature_factor = (df['temp'].to_frame() + 273.15) * 0.082

    # Convert particulate species (μg/m³ → μmol/m³)
    df_particles = concat([
        (df[key] / MOLECULAR_WEIGHTS[key]).copy() for key in particle_keys
    ], axis=1)

    # Convert gaseous species (μg/m³ → ppm)
    df_gases = df[gas_keys] / temperature_factor.values

    # Combine results
    return concat([df_particles, df_gases], axis=1)


def volume_average_mixing(df_volume, df_alwc=None):
    """
    Calculate volume-average refractive index using mixing rule.

    This function calculates the dry and ambient refractive indices
    based on volume-weighted mixing of individual species at 550 nm.

    Parameters
    ----------
    df_volume : DataFrame
        Volume concentration data (μm³/m³) with columns:
        - total_dry : Total dry aerosol volume concentration (required)
        - At least one of: AS_volume, AN_volume, OM_volume, Soil_volume, SS_volume, EC_volume
    df_alwc : DataFrame, optional
        Aerosol liquid water content (μg/m³) with 'ALWC' column.

    Returns
    -------
    DataFrame
        Refractive index data with columns:
        - n_dry : Real part of dry aerosol RI (dimensionless)
        - k_dry : Imaginary part of dry aerosol RI (dimensionless)
        - n_amb : Real part of ambient (wet) aerosol RI
        - k_amb : Imaginary part of ambient aerosol RI
        - gRH : Hygroscopic growth factor (Dp_wet/Dp_dry)

    Raises
    ------
    ValueError
        If required columns are missing.

    Notes
    -----
    Volume-average mixing rule: RI_mix = Σ(Vi * RIi) / V_total
    """
    import numpy as np
    from pandas import DataFrame

    # Validate required columns
    validate_inputs(df_volume, VOLUME_MIXING_REQUIRED, 'volume_average_mixing', VOLUME_COLUMN_DESCRIPTIONS)

    # Check that at least one volume species exists
    available_species = [col for col in VOLUME_MIXING_SPECIES if f'{col}_volume' in df_volume.columns]
    if not available_species:
        volume_cols = [f'{sp}_volume' for sp in VOLUME_MIXING_SPECIES]
        raise ValueError(
            f"\nvolume_average_mixing() 至少需要一個體積欄位！\n"
            f"  可用欄位: {volume_cols}\n"
            f"  現有欄位: {sorted(df_volume.columns.tolist())}"
        )

    if df_alwc is not None:
        validate_inputs(df_alwc, GRH_ALWC_REQUIRED, 'volume_average_mixing (df_alwc)', GRH_COLUMN_DESCRIPTIONS)

    # Refractive index values at 550 nm, sourced from core/_constants.py.
    # Split the complex values into the legacy n / k dicts used below.
    _ri = REFRACTIVE_INDEX['550']
    RI_values = {
        'n': {sp: _ri[sp].real for sp in _ri},
        'k': {sp: _ri[sp].imag for sp in _ri},
    }

    volume_cols = ['AS', 'AN', 'OM', 'Soil', 'SS', 'EC']
    volume_ratio = DataFrame(index=df_volume.index)

    for col in volume_cols:
        if f'{col}_volume' in df_volume.columns:
            volume_ratio[f'{col}_volume_ratio'] = df_volume[f'{col}_volume'] / df_volume['total_dry']

    result = DataFrame(index=df_volume.index)

    result['n_dry'] = sum(
        RI_values['n'][col] * volume_ratio[f'{col}_volume_ratio']
        for col in volume_cols if f'{col}_volume_ratio' in volume_ratio.columns
    )

    result['k_dry'] = sum(
        RI_values['k'][col] * volume_ratio[f'{col}_volume_ratio']
        for col in volume_cols if f'{col}_volume_ratio' in volume_ratio.columns
    )

    if df_alwc is not None and 'ALWC' in df_alwc.columns:
        v_dry = df_volume['total_dry']
        v_wet = df_volume['total_dry'] + df_alwc['ALWC']

        multiplier = v_dry / v_wet
        alwc_ratio = 1 - multiplier

        result['ALWC_volume_ratio'] = alwc_ratio

        result['n_amb'] = (
                sum(
                    RI_values['n'][col] * volume_ratio[f'{col}_volume_ratio']
                    for col in volume_cols if f'{col}_volume_ratio' in volume_ratio.columns
                ) * multiplier +
                RI_values['n']['ALWC'] * alwc_ratio
        )

        result['k_amb'] = (
                sum(
                    RI_values['k'][col] * volume_ratio[f'{col}_volume_ratio']
                    for col in volume_cols if f'{col}_volume_ratio' in volume_ratio.columns
                ) * multiplier
        )

        # Delegate to gRH_calculate() so the formula lives in one place.
        result['gRH'] = gRH_calculate(df_volume, df_alwc)['gRH']
    else:
        result['n_amb'] = np.nan
        result['k_amb'] = np.nan
        result['gRH'] = np.nan

    return result


def kappa_calculate(df_data, diameter=0.5):
    """
    Calculate the hygroscopicity parameter kappa.

    Parameters
    ----------
    df_data : DataFrame
        Data containing:
        - gRH : Hygroscopic growth factor
        - AT  : Ambient temperature (Celsius)
        - RH  : Relative humidity (%)
    diameter : float, default=0.5
        Particle dry diameter in micrometers.

    Returns
    -------
    DataFrame
        Kappa values with 'kappa_chem' column.

    Raises
    ------
    ValueError
        If required columns (gRH, AT, RH) are missing.

    Examples
    --------
    >>> cols = get_required_columns()['kappa_calculate']
    >>> print(cols)
    ['gRH', 'AT', 'RH']
    """
    import numpy as np
    from pandas import DataFrame

    # Validate required columns
    validate_inputs(df_data, KAPPA_REQUIRED, 'kappa_calculate', KAPPA_COLUMN_DESCRIPTIONS)

    surface_tension = 0.072
    Mw = 18
    density = 1
    R = 8.314

    result = DataFrame(index=df_data.index)

    T = df_data['AT'] + 273.15
    A = 4 * (surface_tension * Mw) / (density * R * T)
    power = A / (diameter * 1e-6)

    a_w = (df_data['RH'] / 100) * np.exp(-power)

    gRH = df_data['gRH']
    result['kappa_chem'] = (gRH ** 3 - 1) * (1 - a_w) / a_w

    return result


def gRH_calculate(df_volume, df_alwc):
    """
    Calculate the hygroscopic growth factor gRH.

    Parameters
    ----------
    df_volume : DataFrame
        Volume data with 'total_dry' column.
    df_alwc : DataFrame
        Aerosol liquid water content with 'ALWC' column.

    Returns
    -------
    DataFrame
        Growth factor data with 'gRH' column.

    Raises
    ------
    ValueError
        If required columns are missing.

    Examples
    --------
    >>> cols = get_required_columns()['gRH_calculate']
    >>> print(cols)
    """
    from pandas import DataFrame

    # Validate required columns
    validate_inputs(df_volume, GRH_VOLUME_REQUIRED, 'gRH_calculate (df_volume)', GRH_COLUMN_DESCRIPTIONS)
    validate_inputs(df_alwc, GRH_ALWC_REQUIRED, 'gRH_calculate (df_alwc)', GRH_COLUMN_DESCRIPTIONS)

    result = DataFrame(index=df_volume.index)

    v_dry = df_volume['total_dry']
    v_wet = v_dry + df_alwc['ALWC']

    result['gRH'] = (v_wet / v_dry) ** (1 / 3)

    return result


def get_required_columns():
    """
    Get required column names for calculation functions.

    Returns
    -------
    dict
        Dictionary with function names as keys and required columns as values.

    Examples
    --------
    >>> cols = get_required_columns()
    >>> print(cols['kappa_calculate'])
    ['gRH', 'AT', 'RH']
    """
    return {
        'convert_mass_to_molar_concentration': {
            'required': ['temp'],
            'ionic_species': list(MOLECULAR_WEIGHTS.keys()),
            'description': 'Converts mass to molar concentration. Ionic species are optional.'
        },
        'volume_average_mixing': {
            'required': VOLUME_MIXING_REQUIRED.copy(),
            'species': [f'{sp}_volume' for sp in VOLUME_MIXING_SPECIES],
            'optional': ['ALWC (for ambient RI calculation)']
        },
        'kappa_calculate': KAPPA_REQUIRED.copy(),
        'gRH_calculate': {
            'df_volume': GRH_VOLUME_REQUIRED.copy(),
            'df_alwc': GRH_ALWC_REQUIRED.copy()
        },
        'partition_ratios': {
            'required': PARTITION_REQUIRED.copy(),
            'species': PARTITION_SPECIES.copy(),
            'description': 'Calculate gas-particle partitioning ratios'
        }
    }


# =============================================================================
# Gas-Particle Partitioning Functions
# =============================================================================

def partition_ratios(df_data):
    """
    Calculate gas-particle partitioning ratios.

    Calculates oxidation ratios and equilibrium partitioning coefficients
    to assess the degree of secondary aerosol formation.

    Parameters
    ----------
    df_data : DataFrame
        Data containing particle and gas concentrations (μg/m³).
        Required column: 'temp' (temperature in Celsius)
        Optional species columns (at least one pair needed):
        - SO42-, SO2 : for SOR (Sulfur Oxidation Ratio)
        - NO3-, NO2 : for NOR (Nitrogen Oxidation Ratio)
        - NO3-, NO2, HNO3 : for NOR_2 (complete nitrogen)
        - NH4+, NH3 : for NTR (Nitrogen Transformation Ratio)
        - Cl-, HCl : for chloride partitioning

    Returns
    -------
    DataFrame
        Partitioning ratios with columns:
        - SOR : SO₄²⁻ / (SO₄²⁻ + SO₂) - Sulfur oxidation ratio
        - NOR : NO₃⁻ / (NO₃⁻ + NO₂) - Nitrogen oxidation ratio
        - NOR_2 : (NO₃⁻ + HNO₃) / (NO₃⁻ + NO₂ + HNO₃) - Complete NOR
        - NTR : NH₄⁺ / (NH₄⁺ + NH₃) - Nitrogen transformation ratio
        - epls_SO42- : Same as SOR (epsilon for sulfate)
        - epls_NO3- : NO₃⁻ / (NO₃⁻ + HNO₃) - Nitrate partitioning
        - epls_NH4+ : Same as NTR (epsilon for ammonium)
        - epls_Cl- : Cl⁻ / (Cl⁻ + HCl) - Chloride partitioning

    Notes
    -----
    **Physical Meaning:**

    - **SOR (Sulfur Oxidation Ratio)**: Indicates the degree of SO₂ → SO₄²⁻
      conversion. Higher values suggest more aged/processed aerosols.
      SOR > 0.1 typically indicates secondary sulfate formation.

    - **NOR (Nitrogen Oxidation Ratio)**: Indicates the degree of NOₓ → NO₃⁻
      conversion. Higher values suggest photochemical aging.

    - **NTR (Nitrogen Transformation Ratio)**: Indicates the conversion of
      gaseous NH₃ to particulate NH₄⁺. Related to acid-base neutralization.

    - **Epsilon (ε)**: Equilibrium partitioning coefficient. Represents the
      fraction in particle phase at thermodynamic equilibrium.

    **Interpretation:**
    - Values near 1.0: Nearly complete conversion to particle phase
    - Values near 0.0: Gas phase dominant
    - Values 0.3-0.7: Active gas-particle partitioning

    Examples
    --------
    >>> import pandas as pd
    >>> data = pd.DataFrame({
    ...     'SO42-': [10.0, 15.0],
    ...     'SO2': [5.0, 3.0],
    ...     'NO3-': [8.0, 12.0],
    ...     'NO2': [20.0, 15.0],
    ...     'temp': [25.0, 30.0]
    ... })
    >>> result = partition_ratios(data)
    >>> print(result['SOR'])  # [0.67, 0.83]
    """
    # Validate temperature column exists
    validate_inputs(df_data, PARTITION_REQUIRED, 'partition_ratios', PARTITION_COLUMN_DESCRIPTIONS)

    # Convert to molar concentrations
    df_mol = convert_mass_to_molar_concentration(df_data)

    result = DataFrame(index=df_data.index)

    # Local alias so the existing call sites read the same; the shared
    # helper from core handles division by zero identically.
    safe_ratio = safe_divide

    # SOR: Sulfur Oxidation Ratio
    # SO₄²⁻ / (SO₄²⁻ + SO₂)
    if 'SO42-' in df_mol.columns and 'SO2' in df_mol.columns:
        result['SOR'] = safe_ratio(
            df_mol['SO42-'],
            df_mol['SO42-'] + df_mol['SO2']
        )
        result['epls_SO42-'] = result['SOR']

    # NOR: Nitrogen Oxidation Ratio
    # NO₃⁻ / (NO₃⁻ + NO₂)
    if 'NO3-' in df_mol.columns and 'NO2' in df_mol.columns:
        result['NOR'] = safe_ratio(
            df_mol['NO3-'],
            df_mol['NO3-'] + df_mol['NO2']
        )

    # NOR_2: Complete NOR including HNO3
    # (NO₃⁻ + HNO₃) / (NO₃⁻ + NO₂ + HNO₃)
    if all(col in df_mol.columns for col in ['NO3-', 'NO2', 'HNO3']):
        result['NOR_2'] = safe_ratio(
            df_mol['NO3-'] + df_mol['HNO3'],
            df_mol['NO3-'] + df_mol['NO2'] + df_mol['HNO3']
        )

    # NTR: Nitrogen Transformation Ratio (also called NTR+)
    # NH₄⁺ / (NH₄⁺ + NH₃)
    if 'NH4+' in df_mol.columns and 'NH3' in df_mol.columns:
        result['NTR'] = safe_ratio(
            df_mol['NH4+'],
            df_mol['NH4+'] + df_mol['NH3']
        )
        result['epls_NH4+'] = result['NTR']

    # Epsilon (ε) for NO3: NO₃⁻ / (NO₃⁻ + HNO₃)
    if 'NO3-' in df_mol.columns and 'HNO3' in df_mol.columns:
        result['epls_NO3-'] = safe_ratio(
            df_mol['NO3-'],
            df_mol['NO3-'] + df_mol['HNO3']
        )

    # Epsilon (ε) for Cl: Cl⁻ / (Cl⁻ + HCl)
    if 'Cl-' in df_mol.columns and 'HCl' in df_mol.columns:
        result['epls_Cl-'] = safe_ratio(
            df_mol['Cl-'],
            df_mol['Cl-'] + df_mol['HCl']
        )

    if result.empty:
        raise ValueError(
            "\npartition_ratios() 需要至少一組氣-固物種對！\n"
            f"  可用物種對: {list(PARTITION_SPECIES.keys())}\n"
            f"  現有欄位: {sorted(df_data.columns.tolist())}\n"
            "  例如: SO42- + SO2, NO3- + NO2, NH4+ + NH3"
        )

    return result



"""Unit tests for Aden-Kerker coated-sphere Mie."""
import numpy as np
import pytest

from AeroViz.optical import mie_core_shell, mie_core_shell_sd, mie_lognormal


pytestmark = [pytest.mark.dataprocess, pytest.mark.optical]


# Black-carbon and shell refractive indices typical at 550 nm
BC_RI = complex(1.95, 0.79)
OM_RI = complex(1.55, 0.0)
WAVELENGTH = 550.0


class TestSingleParticle:
    def test_returns_7_keys(self):
        out = mie_core_shell(BC_RI, OM_RI,
                             d_core=50, d_total=150, wavelength=WAVELENGTH)
        for key in ('Q_ext', 'Q_sca', 'Q_abs', 'g', 'Q_pr', 'Q_back', 'Q_ratio'):
            assert key in out
            assert np.isfinite(out[key]), f"{key} is not finite"

    def test_qext_equals_qsca_plus_qabs(self):
        out = mie_core_shell(BC_RI, OM_RI,
                             d_core=50, d_total=150, wavelength=WAVELENGTH)
        np.testing.assert_allclose(
            out['Q_ext'], out['Q_sca'] + out['Q_abs'], rtol=1e-8
        )

    def test_degenerate_no_core_matches_homogeneous(self):
        """d_core=0 → pure shell sphere = homogeneous Mie at d_total."""
        coated = mie_core_shell(BC_RI, OM_RI,
                                d_core=0, d_total=150, wavelength=WAVELENGTH)
        homog = mie_lognormal(
            OM_RI, wavelength=WAVELENGTH,
            geo_mean=150, geo_std=1.001, total_number=1,
        )
        # Both should describe the same sphere; sanity-check Q_abs ≈ 0 since
        # OM is non-absorbing
        assert abs(coated['Q_abs']) < 1e-3

    def test_degenerate_same_ri_matches_homogeneous(self):
        """m_core == m_shell → homogeneous Mie at d_total."""
        out = mie_core_shell(OM_RI, OM_RI,
                             d_core=50, d_total=150, wavelength=WAVELENGTH)
        assert abs(out['Q_abs']) < 1e-3
        assert out['Q_ext'] > 0

    def test_absorption_enhancement(self):
        """BC core coated with OM → E_abs in canonical 1.5-2.5× range."""
        bare = mie_core_shell(BC_RI, BC_RI,
                              d_core=50, d_total=50, wavelength=WAVELENGTH)
        coated = mie_core_shell(BC_RI, OM_RI,
                                d_core=50, d_total=150, wavelength=WAVELENGTH)
        # Scale by cross-section areas to compare per-particle absorption
        area_bare = np.pi * (50e-9) ** 2 / 4
        area_coated = np.pi * (150e-9) ** 2 / 4
        e_abs = (coated['Q_abs'] * area_coated) / (bare['Q_abs'] * area_bare)
        assert 1.0 < e_abs < 4.0, f"E_abs out of range: {e_abs:.2f}"

    def test_negative_diameter_raises(self):
        with pytest.raises(ValueError, match="d_core"):
            mie_core_shell(BC_RI, OM_RI,
                           d_core=-1, d_total=150, wavelength=WAVELENGTH)

    def test_core_larger_than_total_raises(self):
        with pytest.raises(ValueError, match="d_total"):
            mie_core_shell(BC_RI, OM_RI,
                           d_core=200, d_total=150, wavelength=WAVELENGTH)


class TestSizeDistribution:
    def test_psd_integrated_returns_ext_sca_abs(self):
        dp_core = np.array([30, 50, 80, 120])
        dp_total = dp_core + 100  # 100 nm coating
        ndp = np.array([1e3, 5e3, 8e3, 2e3])
        out = mie_core_shell_sd(BC_RI, OM_RI, dp_core, dp_total, ndp,
                                wavelength=WAVELENGTH, psd_type='dN')
        assert {'ext', 'sca', 'abs'}.issubset(out)
        assert out['ext'] > 0
        np.testing.assert_allclose(out['ext'], out['sca'] + out['abs'],
                                   rtol=1e-8)



"""
Mass and volume reconstruction for aerosol chemical composition.

This module reconstructs aerosol mass and volume from ionic species measurements,
handling both ammonium-sufficient and ammonium-deficient conditions.

Required Input Columns
----------------------
- NH4+ : Ammonium (ug/m3)
- SO42-: Sulfate (ug/m3)
- NO3- : Nitrate (ug/m3)
- Fe   : Iron (ug/m3) - for Soil calculation
- Na+  : Sodium (ug/m3) - for Sea Salt calculation
- OC   : Organic Carbon (ug/m3)
- EC   : Elemental Carbon (ug/m3)

Output Species
--------------
- AS   : Ammonium Sulfate (NH4)2SO4
- AN   : Ammonium Nitrate NH4NO3
- OM   : Organic Matter
- Soil : Soil/Crustal matter
- SS   : Sea Salt
- EC   : Elemental Carbon
"""

from pandas import concat, DataFrame

from AeroViz.dataProcess.core import validate_inputs

# =============================================================================
# Constants
# =============================================================================

# Required input columns
REQUIRED_COLUMNS = ['NH4+', 'SO42-', 'NO3-', 'Fe', 'Na+', 'OC', 'EC']

# Input column descriptions 輸入欄位說明
COLUMN_DESCRIPTIONS = {
    'NH4+': 'Particulate Ammonium (μg/m³) 顆粒態銨鹽',
    'SO42-': 'Particulate Sulfate (μg/m³) 顆粒態硫酸鹽',
    'NO3-': 'Particulate Nitrate (μg/m³) 顆粒態硝酸鹽',
    'Fe': 'Iron (μg/m³) 鐵 - 用於計算土壤/地殼物質 Soil',
    'Na+': 'Sodium (μg/m³) 鈉 - 用於計算海鹽 Sea Salt',
    'OC': 'Organic Carbon (μg/m³) 有機碳 - 用於計算有機物 OM',
    'EC': 'Elemental Carbon (μg/m³) 元素碳',
}

# Molecular weights (g/mol)
MOLECULAR_WEIGHT = {
    'NH4+': 18,
    'SO42-': 96,
    'NO3-': 62,
    'AS': 132,  # (NH4)2SO4
    'AN': 80,   # NH4NO3
}

# Conversion: raw species -> reconstructed species
SPECIES_MAPPING = {
    'AS': 'SO42-',
    'AN': 'NO3-',
    'OM': 'OC',
    'Soil': 'Fe',
    'SS': 'Na+',
    'EC': 'EC',
}

# Mass reconstruction coefficients
# AS: (NH4)2SO4 / SO4 = 132/96 = 1.375
# AN: NH4NO3 / NO3 = 80/62 = 1.29
MASS_COEFFICIENTS = {
    'AS': 1.375,
    'AN': 1.29,
    'OM': 1.8,
    'Soil': 28.57,
    'SS': 2.54,
    'EC': 1.0,
}

# Density for volume calculation (g/cm3)
DENSITY_COEFFICIENTS = {
    'AS': 1.76,
    'AN': 1.73,
    'OM': 1.4,
    'Soil': 2.6,
    'SS': 2.16,
    'EC': 1.5,
}

# Refractive index at different wavelengths — see core/_constants.py for
# provenance.
from AeroViz.dataProcess.core import REFRACTIVE_INDEX  # noqa: E402


# =============================================================================
# Helper Functions
# =============================================================================

def calculate_molar_concentrations(df):
    """Calculate molar concentrations from mass concentrations."""
    mol_NH4 = df['NH4+'] / MOLECULAR_WEIGHT['NH4+']
    mol_SO4 = df['SO42-'] / MOLECULAR_WEIGHT['SO42-']
    mol_NO3 = df['NO3-'] / MOLECULAR_WEIGHT['NO3-']
    return mol_NH4, mol_SO4, mol_NO3


def calculate_nh4_status(mol_NH4, mol_SO4, mol_NO3, index):
    """
    Calculate ammonium status (neutralization ratio).

    NH4 status = mol_NH4 / (2 * mol_SO4 + mol_NO3)
    - >= 1: Ammonium sufficient (Enough)
    - < 1: Ammonium deficient (Deficiency)
    """
    ratio = mol_NH4 / (2 * mol_SO4 + mol_NO3)

    df_status = DataFrame(index=index)
    df_status['ratio'] = ratio
    df_status['status'] = ratio.apply(lambda x: 'Enough' if x >= 1 else 'Deficiency')

    return df_status, ratio


def reconstruct_mass_enough(df, mol_NH4, mol_SO4, mol_NO3, oa_oc_ratio=None):
    """
    Reconstruct mass for NH4-sufficient conditions.

    When NH4 is sufficient:
    - AS = SO42- * 1.375 (full neutralization)
    - AN = NO3- * 1.29 (full neutralization)
    - OM = OC * oa_oc_ratio (default 1.8 from MASS_COEFFICIENTS)

    Parameters
    ----------
    oa_oc_ratio : float, optional
        Override the OM/OC conversion ratio. If None, uses
        MASS_COEFFICIENTS['OM'] (= 1.8). Pass an explicit value to keep the
        base reconstruction consistent with the EC-tracer POA/SOA split when
        a non-default oa_oc_ratio is also passed to reconstruction_basic().
    """
    df_mass = DataFrame(index=df.index)

    coefficients = dict(MASS_COEFFICIENTS)
    if oa_oc_ratio is not None:
        coefficients['OM'] = oa_oc_ratio

    for species, coef in coefficients.items():
        raw_col = SPECIES_MAPPING[species]
        df_mass[species] = df[raw_col] * coef

    return df_mass


def adjust_mass_deficiency(df_mass, mol_NH4, mol_SO4, mol_NO3, status_ratio):
    """
    Adjust AS and AN mass for NH4-deficient conditions.

    When NH4 is deficient (ratio < 1):
    1. Calculate residual NH4 after neutralizing SO4: residual = mol_NH4 - 2*mol_SO4
    2. If residual > 0: Some NH4 left to neutralize NO3
       - AN = min(residual, mol_NO3) * 80
    3. If residual <= 0: Not enough NH4 even for SO4
       - AN = 0
       - AS = mol_NH4/2 * 132 (partial neutralization)
    """
    deficient_mask = status_ratio < 1
    if not deficient_mask.any():
        return df_mass

    residual = mol_NH4 - 2 * mol_SO4

    # Case 1: residual > 0 (some NH4 left for NO3)
    pos_residual = residual > 0
    if pos_residual.any():
        # AN limited by residual or available NO3
        cond = pos_residual & (residual <= mol_NO3)
        df_mass.loc[cond, 'AN'] = residual.loc[cond] * MOLECULAR_WEIGHT['AN']

        cond = pos_residual & (residual > mol_NO3)
        df_mass.loc[cond, 'AN'] = mol_NO3.loc[cond] * MOLECULAR_WEIGHT['AN']

    # Case 2: residual <= 0 (not enough NH4 for SO4)
    neg_residual = residual <= 0
    if neg_residual.any():
        df_mass.loc[neg_residual, 'AN'] = 0

        # Partial AS neutralization
        cond = neg_residual & (mol_NH4 <= 2 * mol_SO4)
        df_mass.loc[cond, 'AS'] = mol_NH4.loc[cond] / 2 * MOLECULAR_WEIGHT['AS']

        cond = neg_residual & (mol_NH4 > 2 * mol_SO4)
        df_mass.loc[cond, 'AS'] = mol_SO4.loc[cond] * MOLECULAR_WEIGHT['AS']

    return df_mass


def calculate_volume(df_mass, df_water=None):
    """
    Calculate species volume concentrations from mass using density coefficients.

    Output columns:
    - {species}_volume: Volume concentration for each species (μm³/m³)
    - total_dry: Total dry aerosol volume concentration (μm³/m³)
    - ALWC: Aerosol liquid water content volume (μm³/m³), if df_water provided
    - total_wet: Total wet aerosol volume (μm³/m³), if df_water provided
    """
    df_vol = DataFrame(index=df_mass.index)

    # Calculate dry volumes (μg/m³ / g/cm³ = μm³/m³)
    for species, density in DENSITY_COEFFICIENTS.items():
        if species in df_mass.columns:
            df_vol[f'{species}_volume'] = df_mass[species] / density

    # Total dry aerosol volume concentration
    volume_cols = [f'{sp}_volume' for sp in DENSITY_COEFFICIENTS.keys() if f'{sp}_volume' in df_vol.columns]
    df_vol['total_dry'] = df_vol[volume_cols].sum(axis=1, min_count=6)

    # Add ALWC (Aerosol Liquid Water Content) if provided
    if df_water is not None:
        df_vol['ALWC'] = df_water.copy()
        df_vol = df_vol.dropna()
        df_vol['total_wet'] = df_vol['total_dry'] + df_vol['ALWC']

    return df_vol


def calculate_refractive_index(df_vol, df_water=None):
    """
    Calculate volume-weighted refractive index at 550nm and 450nm.

    Output:
    - RI_dry: Dry aerosol refractive index (complex: n + kj)
    - RI_wet: Wet aerosol refractive index (if ALWC provided)
    """
    ri_results = {}

    for wavelength, ri_coef in REFRACTIVE_INDEX.items():
        df_ri = DataFrame(index=df_vol.index)

        # Calculate RI contribution from each species (volume * RI)
        available_species = []
        for species in DENSITY_COEFFICIENTS.keys():
            vol_col = f'{species}_volume'
            if vol_col in df_vol.columns:
                df_ri[species] = df_vol[vol_col] * ri_coef[species]
                available_species.append(species)

        # Dry RI (volume-weighted average): sum(Vi * RIi) / total_V
        df_ri['RI_dry'] = (df_ri[available_species] / df_vol['total_dry'].values.reshape(-1, 1)).sum(axis=1)

        # Wet RI (if ALWC provided)
        df_ri['RI_wet'] = None
        if df_water is not None and 'total_wet' in df_vol.columns:
            df_ri['ALWC'] = df_vol['ALWC'] * ri_coef['ALWC']
            all_species = available_species + ['ALWC']
            df_ri['RI_wet'] = (df_ri[all_species] / df_vol['total_wet'].values.reshape(-1, 1)).sum(axis=1)

        ri_results[f'RI_{wavelength}'] = df_ri[['RI_dry', 'RI_wet']]

    return ri_results


def calculate_density(df_mass, df_vol, df_all, df_density=None):
    """Calculate aerosol density (reconstructed and measured)."""
    # Reconstructed density
    density_rec = df_mass['total'] / df_vol['total_dry']

    # Measured density (if density data provided)
    if df_density is not None:
        df_den_all = concat([
            df_all[['SO42-', 'NO3-', 'NH4+', 'EC']],
            df_density,
            df_mass['OM']
        ], axis=1).dropna()

        vol_cal = (
            df_den_all[['SO42-', 'NO3-', 'NH4+']].sum(axis=1) / 1.75 +
            df_den_all['Cl-'] / 1.52 +
            df_den_all['OM'] / 1.4 +
            df_den_all['EC'] / 1.77
        )
        density_mat = df_den_all.sum(axis=1, min_count=6) / vol_cal
    else:
        vol_cal = DataFrame()
        density_mat = density_rec

    return density_mat, density_rec, vol_cal


def calculate_equivalents(mol_NH4, mol_SO4, mol_NO3):
    """Calculate molar and equivalent concentrations."""
    df_eq = concat([
        mol_NH4, mol_SO4, mol_NO3,
        mol_NH4 * 1,  # eq_NH4 (charge = 1)
        mol_SO4 * 2,  # eq_SO4 (charge = 2)
        mol_NO3 * 1   # eq_NO3 (charge = 1)
    ], axis=1)
    df_eq.columns = ['mol_NH4', 'mol_SO4', 'mol_NO3', 'eq_NH4', 'eq_SO4', 'eq_NO3']
    return df_eq


# =============================================================================
# EC-Tracer Method for POA/SOA Split
# =============================================================================

def _split_om_ec_tracer(df_mass, df_all, oa_oc_ratio=1.8):
    """
    Split OM into POA and SOA using the EC-tracer (Minimum R Squared) method.

    The method finds the primary OC/EC ratio by minimizing R² between
    SOC and EC, ensuring they are independent (Lim & Turpin, 2002).

    POC = (OC/EC)_primary × EC
    SOC = OC - POC
    POA = POC × oa_oc_ratio
    SOA = SOC × oa_oc_ratio

    Parameters
    ----------
    df_mass : DataFrame
        Reconstructed mass with 'OM' column.
    df_all : DataFrame
        Raw input data with 'OC' and 'EC' columns.
    oa_oc_ratio : float
        OA/OC conversion ratio.

    Returns
    -------
    DataFrame
        df_mass with added POA, SOA, POC, SOC columns.
    """
    import numpy as np
    from AeroViz.dataProcess.Chemistry._ocec import find_mrs_ratio

    oc = df_all['OC']
    ec = df_all['EC']

    # Use the canonical MRS implementation in _ocec.py so the OC/EC
    # tracer search is identical to the one used by the OCEC processing
    # pipeline. Scan range follows the convention used previously:
    # from 0.5 up to the 95th-percentile of (OC/EC) in the data.
    valid = oc.notna() & ec.notna() & (ec > 0)
    if not valid.any():
        df_mass['POC'] = np.nan
        df_mass['SOC'] = np.nan
        df_mass['POA'] = np.nan
        df_mass['SOA'] = np.nan
        return df_mass

    oc_ec_ratios = np.arange(0.5, (oc[valid] / ec[valid]).quantile(0.95), 0.01)
    best_ratio, _ = find_mrs_ratio(oc, ec, oc_ec_ratios)

    # Calculate POC and SOC (clip negatives to enforce physical bounds on
    # the output; the search itself uses unclipped SOC, matching canonical
    # MRS).
    df_mass['POC'] = (best_ratio * ec).clip(lower=0)
    df_mass['SOC'] = (oc - df_mass['POC']).clip(lower=0)
    df_mass['POA'] = df_mass['POC'] * oa_oc_ratio
    df_mass['SOA'] = df_mass['SOC'] * oa_oc_ratio

    return df_mass


# =============================================================================
# Main Function
# =============================================================================

def reconstruction_basic(df_che, df_ref, df_water=None, df_density=None, nam_lst=None,
                         split_om=False, oa_oc_ratio=1.8):
    """
    Reconstruct aerosol mass and volume from chemical composition.

    This function converts ionic species (NH4+, SO42-, NO3-, etc.) to
    reconstructed species (AS, AN, OM, Soil, SS, EC) considering the
    ammonium neutralization status.

    Parameters
    ----------
    df_che : tuple of DataFrames
        Chemical composition data. Will be concatenated and renamed to nam_lst.
    df_ref : DataFrame or Series
        Reference mass (e.g., PM2.5) for quality control.
    df_water : DataFrame or None, optional
        Aerosol liquid water content (ALWC).
    df_density : DataFrame or None, optional
        Measured density data (requires 'Cl-' column).
    nam_lst : list, optional
        Column names for df_che after concatenation.
        Default: ['NH4+', 'SO42-', 'NO3-', 'Fe', 'Na+', 'OC', 'EC']
    split_om : bool, optional
        If True, split OM into POA and SOA using the EC-tracer method.
        Requires 'OC' and 'EC' columns. Default: False.
    oa_oc_ratio : float, optional
        OA/OC conversion ratio for POA/SOA calculation. Default: 1.8.

    Returns
    -------
    dict
        Dictionary containing:
        - 'mass': Reconstructed mass (AS, AN, OM, Soil, SS, EC, total)
                  If split_om=True, also includes POA, SOA, POC, SOC columns.
        - 'volume': Reconstructed volume (species + total_dry, total_wet)
        - 'vol_cal': Calculated volume for density
        - 'eq': Molar and equivalent concentrations
        - 'NH4_status': Ammonium status ('ratio' and 'status')
        - 'density_mat': Measured density
        - 'density_rec': Reconstructed density
        - 'RI_550': Refractive index at 550nm
        - 'RI_450': Refractive index at 450nm

    Raises
    ------
    ValueError
        If required columns are missing.

    Examples
    --------
    >>> result = reconstruction_basic(
    ...     df_che=(df_ions, df_carbon),
    ...     df_ref=df_pm25,
    ...     df_water=df_alwc,
    ...     nam_lst=['NH4+', 'SO42-', 'NO3-', 'Fe', 'Na+', 'OC', 'EC'],
    ...     split_om=True
    ... )
    >>> result['mass']       # Includes POA, SOA columns
    >>> result['mass'][['POA', 'SOA']].sum(axis=1)  # Should equal OM
    """
    # Default column names
    if nam_lst is None:
        nam_lst = REQUIRED_COLUMNS

    # Prepare input data
    df_all = concat(df_che, axis=1)
    original_index = df_all.index.copy()
    df_all.columns = nam_lst

    # Validate required columns
    validate_inputs(df_all, REQUIRED_COLUMNS, 'reconstruction_basic', COLUMN_DESCRIPTIONS)

    # Step 1: Calculate molar concentrations
    mol_NH4, mol_SO4, mol_NO3 = calculate_molar_concentrations(df_all)

    # Step 2: Calculate NH4 status
    df_nh4_status, status_ratio = calculate_nh4_status(mol_NH4, mol_SO4, mol_NO3, original_index)

    # Step 3: Reconstruct mass (assuming NH4 sufficient).
    # Pass `oa_oc_ratio` so the base OM = OC × ratio is consistent with the
    # split_om branch below; passing None keeps the default 1.8.
    df_mass = reconstruct_mass_enough(
        df_all, mol_NH4, mol_SO4, mol_NO3,
        oa_oc_ratio=oa_oc_ratio,
    )

    # Step 4: Adjust for NH4 deficiency
    df_mass = adjust_mass_deficiency(df_mass, mol_NH4, mol_SO4, mol_NO3, status_ratio)

    # Step 4.5: Split OM into POA and SOA using EC-tracer (MRS) method
    if split_om:
        df_mass = _split_om_ec_tracer(df_mass, df_all, oa_oc_ratio)

    df_mass['total'] = df_mass.sum(axis=1, min_count=6)

    # Step 5: Calculate volume
    df_mass_valid = df_mass.dropna().copy()
    df_vol = calculate_volume(df_mass_valid, df_water)

    # Step 6: Calculate density
    density_mat, density_rec, vol_cal = calculate_density(df_mass, df_vol, df_all, df_density)

    # Step 7: Calculate refractive index
    ri_results = calculate_refractive_index(df_vol, df_water)

    # Step 8: Calculate equivalents
    df_eq = calculate_equivalents(mol_NH4, mol_SO4, mol_NO3)

    # Compile output
    out = {
        'mass': df_mass,
        'volume': df_vol,
        'vol_cal': vol_cal,
        'eq': df_eq,
        'NH4_status': df_nh4_status,
        'density_mat': density_mat,
        'density_rec': density_rec,
    }
    out.update(ri_results)

    # Reindex all outputs to original index
    for key, value in out.items():
        if hasattr(value, 'reindex'):
            out[key] = value.reindex(original_index)

    return out


# =============================================================================
# Utility Functions
# =============================================================================

def mass_ratio(df):
    """
    Calculate mass ratios relative to PM2.5.

    Parameters
    ----------
    df : Series
        Must contain 'PM25' and 'total_mass' values.

    Returns
    -------
    Series
        Mass ratios for each species.
    """
    if df['PM25'] >= df['total_mass']:
        df['others'] = df['PM25'] - df['total_mass']
    else:
        df['others'] = 0

    for val, species in zip(df.values, df.index):
        df[f'{species}_ratio'] = round(val / df['PM25'], 3)

    return df['others':].drop(labels=['PM25_ratio', 'total_mass_ratio'])


def get_required_columns():
    """
    Get required column names and output descriptions.

    Returns
    -------
    dict
        Documentation for reconstruction_basic inputs and outputs.
    """
    return {
        'reconstruction_basic': {
            'required_columns': REQUIRED_COLUMNS.copy(),
            'column_descriptions': COLUMN_DESCRIPTIONS.copy(),
            'outputs': {
                'mass': 'Reconstructed mass (AS, AN, OM, Soil, SS, EC, total)',
                'volume': 'Reconstructed volume with ALWC',
                'eq': 'Molar and equivalent concentrations',
                'NH4_status': "Ammonium status: 'ratio' and 'status' (Enough/Deficiency)",
                'density_mat': 'Measured density',
                'density_rec': 'Reconstructed density',
                'RI_550': 'Refractive index at 550nm (RI_dry, RI_wet)',
                'RI_450': 'Refractive index at 450nm (RI_dry, RI_wet)',
            },
            'coefficients': {
                'mass': MASS_COEFFICIENTS,
                'density': DENSITY_COEFFICIENTS,
            }
        },
    }


# Backward compatibility
_basic = reconstruction_basic
DEFAULT_REQUIRED_COLUMNS = REQUIRED_COLUMNS



# -*- coding: utf-8 -*-
"""
Mie Scattering Calculation for Size Distribution Data

This module provides vectorized Mie scattering calculations optimized for
particle size distribution (PSD) data stored in pandas DataFrames.

Based on: http://pymiescatt.readthedocs.io/en/latest/forward.html

Theory:
    Mie theory describes the scattering of electromagnetic radiation by
    spherical particles. The key outputs are:
    - Q_ext: Extinction efficiency (scattering + absorption)
    - Q_sca: Scattering efficiency
    - Q_abs: Absorption efficiency (Q_ext - Q_sca)
"""

import warnings
import numpy as np
import pandas as pd
from scipy.integrate import trapezoid
from scipy.special import jv, yv  # Bessel functions


# =============================================================================
# PSD Type Detection and Integration
# =============================================================================

def _detect_psd_type(values: np.ndarray, diameter: np.ndarray) -> tuple[str, str]:
    """
    Auto-detect whether PSD data is dN/dlogDp or dN.

    Parameters
    ----------
    values : np.ndarray
        PSD values, shape (n_bins,) or (n_times, n_bins)
    diameter : np.ndarray
        Particle diameters in nm

    Returns
    -------
    psd_type : str
        'dNdlogDp' or 'dN'
    confidence : str
        'high', 'medium', or 'low'
    """
    log_dp = np.log10(diameter)
    dlogdp = np.diff(log_dp).mean()

    # Use mean values if 2D
    if values.ndim > 1:
        values_1d = np.nanmean(values, axis=0)
    else:
        values_1d = values

    # Calculate total N under both assumptions
    N_as_dNdlogDp = trapezoid(values_1d, x=log_dp)
    N_as_dN = np.nansum(values_1d)

    # Typical total particle number: 1e2 - 1e7 #/cm³
    typical_min, typical_max = 1e2, 1e7

    dNdlogDp_ok = typical_min <= N_as_dNdlogDp <= typical_max
    dN_ok = typical_min <= N_as_dN <= typical_max

    # Calculate the ratio (should be ~1/dlogDp if dNdlogDp)
    ratio = N_as_dN / N_as_dNdlogDp if N_as_dNdlogDp > 0 else float('inf')
    expected_ratio = 1 / dlogdp

    if dNdlogDp_ok and not dN_ok:
        return 'dNdlogDp', 'high'
    elif dN_ok and not dNdlogDp_ok:
        return 'dN', 'high'
    elif dNdlogDp_ok and dN_ok:
        # Both reasonable - check if ratio matches expected
        if 0.5 * expected_ratio < ratio < 2 * expected_ratio:
            return 'dNdlogDp', 'medium'
        else:
            return 'dN', 'medium'
    else:
        # Neither reasonable - default to dNdlogDp with warning
        return 'dNdlogDp', 'low'


def _integrate_psd(
    values: np.ndarray,
    diameter: np.ndarray,
    psd_type: str = 'dNdlogDp'
) -> np.ndarray:
    """
    Integrate over particle size distribution.

    Parameters
    ----------
    values : np.ndarray
        Integrand values, shape (n_times, n_bins)
    diameter : np.ndarray
        Particle diameters in nm
    psd_type : str
        'dNdlogDp' or 'dN'

    Returns
    -------
    result : np.ndarray
        Integrated values, shape (n_times,)
    """
    if psd_type == 'dNdlogDp':
        log_dp = np.log10(diameter)
        return trapezoid(values, x=log_dp, axis=-1)
    else:  # dN
        return np.sum(values, axis=-1)


def calculate_mie_coefficients(
    refractive_index: np.ndarray,
    size_parameter: np.ndarray,
    n_max: np.ndarray,
    n_terms: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Calculate Mie scattering coefficients (a_n, b_n) for multiple particles.

    This implements the core Mie theory calculation using Bessel functions
    and the logarithmic derivative method for numerical stability.

    Parameters
    ----------
    refractive_index : np.ndarray
        Complex refractive index (m = n + ik) for each time point.
        Shape: (n_times,)
    size_parameter : np.ndarray
        Size parameter x = π * diameter / wavelength for each size bin.
        Shape: (n_bins,)
    n_max : np.ndarray
        Maximum number of terms needed for each size bin.
        Shape: (n_bins,)
    n_terms : pd.DataFrame
        Term indices for the series expansion.
        Shape: (n_bins, max_terms)

    Returns
    -------
    Q_ext : pd.DataFrame
        Extinction efficiency for each (size_bin, time_point).
    Q_sca : pd.DataFrame
        Scattering efficiency for each (size_bin, time_point).

    Notes
    -----
    The Mie coefficients a_n and b_n are calculated using:
    - Riccati-Bessel functions (ψ, χ)
    - Logarithmic derivative D_n(mx) computed via downward recurrence
    """
    m = refractive_index
    x = size_parameter
    n_bins = len(x)
    n_times = len(m)

    # Bessel function order: ν = n + 0.5
    nu = n_terms.copy() + 0.5

    # Coefficient for series summation: 2n + 1
    coeff_2n_plus_1 = 2 * n_terms.copy() + 1

    # === Calculate Riccati-Bessel functions ===
    # ψ_n(x) = sqrt(πx/2) * J_{n+1/2}(x)  [Bessel J]
    # χ_n(x) = -sqrt(πx/2) * Y_{n+1/2}(x) [Bessel Y]
    sqrt_factor = np.sqrt(0.5 * np.pi * x)

    psi_n = sqrt_factor.reshape(-1, 1) * jv(nu, x.reshape(-1, 1))
    chi_n = -sqrt_factor.reshape(-1, 1) * yv(nu, x.reshape(-1, 1))

    # ψ_{n-1}(x) and χ_{n-1}(x) with boundary conditions
    psi_n_minus_1 = pd.concat(
        [pd.DataFrame(np.sin(x)), psi_n.mask(n_terms == n_max.reshape(-1, 1))],
        axis=1
    )
    psi_n_minus_1.columns = np.arange(len(psi_n_minus_1.columns))
    psi_n_minus_1 = psi_n_minus_1[n_terms.columns]

    chi_n_minus_1 = pd.concat(
        [pd.DataFrame(np.cos(x)), chi_n.mask(n_terms == n_max.reshape(-1, 1))],
        axis=1
    )
    chi_n_minus_1.columns = np.arange(len(chi_n_minus_1.columns))
    chi_n_minus_1 = chi_n_minus_1[n_terms.columns]

    # Hankel function: ξ_n(x) = ψ_n(x) - i*χ_n(x)
    xi_n = psi_n - 1j * chi_n
    xi_n_minus_1 = psi_n_minus_1 - 1j * chi_n_minus_1

    # === Calculate logarithmic derivative D_n(mx) ===
    mx = m.reshape(-1, 1) * x  # Complex argument

    # Number of terms needed for downward recurrence
    nmx_array = np.round(
        np.max(
            np.hstack([[n_max] * n_times, np.abs(mx)]).reshape(n_times, 2, -1),
            axis=1
        ) + 16
    )

    # Initialize output DataFrames
    Q_ext = pd.DataFrame(columns=m.flatten(), index=n_terms.index)
    Q_sca = pd.DataFrame(columns=m.flatten(), index=n_terms.index)
    g_raw = pd.DataFrame(columns=m.flatten(), index=n_terms.index)
    # Complex backscatter coherent sum: |Σ (2n+1)·(−1)ⁿ·(aₙ−bₙ)|² / x² = Q_back
    back_raw = pd.DataFrame(columns=m.flatten(), index=n_terms.index, dtype=complex)

    # Normalize n/x for later use
    n_over_x = n_terms / x.reshape(-1, 1)

    # Pre-compute g and Q_back angular-series coefficients (functions of n only).
    n_values = n_terms.values  # shape (n_bins, max_terms)
    coef_g_a = n_values * (n_values + 2) / (n_values + 1)         # n(n+2)/(n+1)
    coef_g_b = (2 * n_values + 1) / (n_values * (n_values + 1))   # (2n+1)/(n(n+1))
    # (−1)ⁿ — NaN-safe: integer where n is valid, NaN otherwise → NaN propagates
    sign_n = (-1.0) ** n_values

    # === Main calculation loop over size bins ===
    for bin_idx, (nmx_values, mx_values, nmax_bin) in enumerate(
        zip(nmx_array.T, mx.T, n_max)
    ):
        # Logarithmic derivative D_n(mx) via downward recurrence
        D_n = pd.DataFrame(
            np.nan,
            index=np.arange(n_times),
            columns=n_terms.columns,
            dtype=complex
        )

        # Group by nmx value for efficient computation
        for nmx, time_indices in pd.DataFrame(nmx_values).groupby(0).groups.items():
            inv_mx = 1 / mx_values[time_indices]
            nmx_int = int(nmx)

            # Downward recurrence: D_{n-1} = n/mx - 1/(D_n + n/mx)
            D_recurrence = np.zeros((len(time_indices), nmx_int), dtype=complex)
            for idx in range(nmx_int - 1, 1, -1):
                D_recurrence[:, idx - 1] = (
                    idx * inv_mx - 1 / (D_recurrence[:, idx] + idx * inv_mx)
                )

            D_n.loc[time_indices, 0:int(nmax_bin) - 1] = D_recurrence[:, 1:int(nmax_bin) + 1]

        # Get values for this size bin
        n_x = n_over_x.loc[bin_idx]
        psi = psi_n.loc[bin_idx]
        psi_prev = psi_n_minus_1.loc[bin_idx]
        xi = xi_n.loc[bin_idx]
        xi_prev = xi_n_minus_1.loc[bin_idx]
        coeff = coeff_2n_plus_1.loc[bin_idx].values

        # === Calculate Mie coefficients a_n and b_n ===
        # a_n = (D_n/m + n/x) * ψ_n - ψ_{n-1}
        #       ─────────────────────────────────
        #       (D_n/m + n/x) * ξ_n - ξ_{n-1}
        numerator_a = D_n / m.reshape(-1, 1) + n_x
        a_n = (numerator_a * psi - psi_prev) / (numerator_a * xi - xi_prev)

        # b_n = (m*D_n + n/x) * ψ_n - ψ_{n-1}
        #       ─────────────────────────────────
        #       (m*D_n + n/x) * ξ_n - ξ_{n-1}
        numerator_b = D_n * m.reshape(-1, 1) + n_x
        b_n = (numerator_b * psi - psi_prev) / (numerator_b * xi - xi_prev)

        # === Calculate efficiencies ===
        # Q_ext = (2/x²) * Σ (2n+1) * Re(a_n + b_n)
        # Q_sca = (2/x²) * Σ (2n+1) * (|a_n|² + |b_n|²)
        real_a, real_b = np.real(a_n), np.real(b_n)
        imag_a, imag_b = np.imag(a_n), np.imag(b_n)

        Q_ext.loc[bin_idx] = np.nansum(coeff * (real_a + real_b), axis=1)
        Q_sca.loc[bin_idx] = np.nansum(
            coeff * (real_a**2 + real_b**2 + imag_a**2 + imag_b**2),
            axis=1
        )

        # === Asymmetry parameter g (vectorised) ===
        # g · Q_sca · x² / 4 = Σ [n(n+2)/(n+1)] · Re(aₙ·aₙ₊₁* + bₙ·bₙ₊₁*)
        #                   + Σ [(2n+1)/(n(n+1))] · Re(aₙ·bₙ*)
        # We compute the bracketed sum here ("g_raw"); the prefactor lives in
        # calculate_mie_efficiencies so g = (4 / (Q_sca · x²)) · g_raw.
        a_arr = np.asarray(a_n)                   # (n_times, max_terms) complex
        b_arr = np.asarray(b_n)
        a_shift = np.concatenate(
            [a_arr[:, 1:], np.full((a_arr.shape[0], 1), np.nan + 0j)], axis=1
        )
        b_shift = np.concatenate(
            [b_arr[:, 1:], np.full((b_arr.shape[0], 1), np.nan + 0j)], axis=1
        )
        # Re(aₙ · conj(aₙ₊₁)) = Re(aₙ)·Re(aₙ₊₁) + Im(aₙ)·Im(aₙ₊₁)
        cross_a = a_arr.real * a_shift.real + a_arr.imag * a_shift.imag
        cross_b = b_arr.real * b_shift.real + b_arr.imag * b_shift.imag
        # Re(aₙ · conj(bₙ))
        ab_cross = a_arr.real * b_arr.real + a_arr.imag * b_arr.imag
        g_raw.loc[bin_idx] = np.nansum(
            coef_g_a[bin_idx] * (cross_a + cross_b)
            + coef_g_b[bin_idx] * ab_cross,
            axis=1,
        )

        # === Backscatter coherent sum (complex) ===
        # Σ (2n+1) · (−1)ⁿ · (aₙ − bₙ) — Q_back = |sum|² / x² (in efficiencies).
        a_minus_b = a_arr - b_arr                            # complex (n_times, max_terms)
        weighted = (coeff * sign_n[bin_idx]) * a_minus_b     # broadcasts (max_terms,) over rows
        back_raw.loc[bin_idx] = np.nansum(weighted, axis=1)

    return Q_ext, Q_sca, g_raw, back_raw


def calculate_mie_efficiencies(
    refractive_index,
    wavelength: float,
    diameter: np.ndarray,
) -> dict:
    """
    Calculate the full set of Mie efficiencies for a particle distribution.

    Returns the seven standard outputs of Mie theory in a single dict.
    The first three (``Q_ext``, ``Q_sca``, ``Q_abs``) come from the fast
    vectorised series in :func:`calculate_mie_coefficients`; the
    remaining four (``g``, ``Q_pr``, ``Q_back``, ``Q_ratio``) are
    obtained per ``(time, bin)`` pair from the JIT-compiled
    :func:`mie_kernels.AutoMieQ` (Phase 7 will collapse this into the
    vectorised loop).

    Parameters
    ----------
    refractive_index : complex | np.ndarray
        Complex refractive index. Either a single scalar (one material
        applied to every bin) or a 1-D array of length ``n_times``
        (one ``m`` per time point). A 1-element numpy array counts as
        the vectorised case — output keeps the first axis with size 1.
    wavelength : float
        Wavelength of incident light in nm.
    diameter : np.ndarray
        Particle diameters in nm. Shape ``(n_bins,)``.

    Returns
    -------
    dict
        Keys (each value is an ndarray):

        * ``Q_ext``   — extinction efficiency
        * ``Q_sca``   — scattering efficiency
        * ``Q_abs``   — absorption efficiency (= Q_ext − Q_sca)
        * ``g``       — asymmetry parameter (forward-scattering bias)
        * ``Q_pr``    — radiation-pressure efficiency (= Q_ext − g·Q_sca)
        * ``Q_back``  — backscatter efficiency
        * ``Q_ratio`` — backscatter ratio (= Q_back / Q_sca)

        Each array has shape ``(n_bins,)`` for a scalar ``m`` and
        ``(n_times, n_bins)`` otherwise.

    Notes
    -----
    Size parameter: ``x = π · d / λ``. Number of terms scales as
    ``n_max ≈ 2 + x + 4·x^(1/3)``.
    """
    # Detect scalar input so we can squeeze it back at the end.
    scalar_input = (
        np.isscalar(refractive_index)
        or isinstance(refractive_index, complex)
        or (isinstance(refractive_index, np.ndarray) and refractive_index.ndim == 0)
    )
    ri_array = np.atleast_1d(np.asarray(refractive_index, dtype=complex))
    diameter = np.asarray(diameter, dtype=float)

    # === Fully vectorised path: all four raw sums in one inner-loop pass ===
    size_parameter = np.pi * diameter / wavelength
    n_max = np.round(2 + size_parameter + 4 * size_parameter ** (1 / 3))
    max_terms = int(n_max.max())
    n_terms = pd.DataFrame([np.arange(1, max_terms + 1)] * len(n_max))
    n_terms = n_terms.mask(n_terms > n_max.reshape(-1, 1))

    Q_ext_raw, Q_sca_raw, g_raw, back_raw = calculate_mie_coefficients(
        ri_array, size_parameter, n_max, n_terms
    )

    # Shape transforms: pandas DataFrame indexed by bin (rows) × time (cols)
    # → numpy array of shape (n_times, n_bins).
    x2 = size_parameter ** 2
    inv_x2 = (1 / x2).reshape(-1, 1)
    Q_ext = (2 * inv_x2 * Q_ext_raw).values.T.astype(float)
    Q_sca = (2 * inv_x2 * Q_sca_raw).values.T.astype(float)
    Q_abs = Q_ext - Q_sca

    # g = (4 / (Q_sca · x²)) · g_raw, with safe divide when Q_sca → 0.
    # Numerator/denominator both still indexed as bin × time:
    g_numer = (4 * inv_x2 * g_raw).values.T.astype(float)   # (n_times, n_bins)
    with np.errstate(invalid='ignore', divide='ignore'):
        g_arr = np.where(Q_sca > 0, g_numer / Q_sca, 0.0)

    # Q_back = |back_raw|² / x²
    back_complex = back_raw.values.T            # (n_times, n_bins)
    Q_back_arr = (np.abs(back_complex) ** 2) * (1 / x2)

    # Q_pr  = Q_ext − g · Q_sca
    Q_pr_arr = Q_ext - g_arr * Q_sca

    # Q_ratio = Q_back / Q_sca (safe divide)
    with np.errstate(invalid='ignore', divide='ignore'):
        Q_ratio_arr = np.where(Q_sca > 0, Q_back_arr / Q_sca, 0.0)

    out = {
        'Q_ext': Q_ext, 'Q_sca': Q_sca, 'Q_abs': Q_abs,
        'g': g_arr, 'Q_pr': Q_pr_arr,
        'Q_back': Q_back_arr, 'Q_ratio': Q_ratio_arr,
    }

    if scalar_input:
        return {k: v[0] for k, v in out.items()}
    return out


def Mie_SD(
    refractive_index: np.ndarray,
    wavelength: float,
    psd: pd.DataFrame,
    psd_type: str = 'auto',
    multi_ri_per_psd: bool = False,
    precomputed_Q: dict = None
) -> pd.DataFrame | dict:
    """
    Calculate optical properties from particle size distribution using Mie theory.

    This function integrates Mie efficiencies over the particle size distribution
    to obtain bulk optical properties (extinction, scattering, absorption).

    Parameters
    ----------
    refractive_index : np.ndarray
        Complex refractive index (m = n + ik).
        - If multi_ri_per_psd=False: Shape (n_times,), one m per time point
        - If multi_ri_per_psd=True: Shape (n_ri,), multiple m tested per PSD
    wavelength : float
        Wavelength of incident light in nm.
    psd : pd.DataFrame
        Particle size distribution data.
        - Columns: particle diameters (nm)
        - Rows: time points
        - Values: dN/dlogDp or dN depending on psd_type
    psd_type : str, default='auto'
        Type of PSD input:
        - 'dNdlogDp': Number concentration per log bin width (#/cm³)
        - 'dN': Number concentration per bin (#/cm³/bin)
        - 'auto': Auto-detect with warning if uncertain
    multi_ri_per_psd : bool, default=False
        If True, calculate for multiple refractive indices per PSD row.
        Useful for refractive index retrieval.
    precomputed_Q : dict, optional
        Pre-computed dict from ``calculate_mie_efficiencies`` (only
        ``Q_ext`` and ``Q_sca`` are read here) to avoid recalculation.

    Returns
    -------
    pd.DataFrame or dict
        If multi_ri_per_psd=False:
            DataFrame with columns ['ext', 'sca', 'abs'] in Mm⁻¹
        If multi_ri_per_psd=True:
            dict with keys 'ext', 'sca', 'abs', each a DataFrame
            with refractive indices as columns

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>>
    >>> # Create sample PSD data (100 time points, 50 size bins)
    >>> dp = np.logspace(1, 3, 50)  # 10-1000 nm
    >>> psd = pd.DataFrame(np.random.rand(100, 50) * 1000, columns=dp)
    >>>
    >>> # Refractive index for each time point
    >>> m = np.array([complex(1.5, 0.02)] * 100)
    >>>
    >>> # Calculate optical properties (explicit dN/dlogDp input)
    >>> result = Mie_SD(m, wavelength=550, psd=psd, psd_type='dNdlogDp')
    >>> print(result[['ext', 'sca', 'abs']].head())

    Notes
    -----
    The optical coefficients are calculated as:

    For dN/dlogDp input:
        b = ∫ Q(Dp) * π/4 * Dp² * (dN/dlogDp) * dlogDp

    For dN input:
        b = Σ Q(Dp) * π/4 * Dp² * dN

    Where:
        - Q: Mie efficiency (extinction, scattering, or absorption)
        - Dp: particle diameter
        - The factor 1e-6 converts from nm² to Mm⁻¹
    """
    # Ensure psd is a DataFrame
    if not isinstance(psd, pd.DataFrame):
        psd = pd.DataFrame(psd).T

    # Validate input dimensions
    if not multi_ri_per_psd and len(refractive_index) != len(psd):
        raise ValueError(
            f"Refractive index array length ({len(refractive_index)}) must match "
            f"PSD row count ({len(psd)}). Set multi_ri_per_psd=True for RI retrieval."
        )

    # Extract diameter and number concentration
    diameter = psd.columns.values.astype(float)  # nm
    number_conc = psd.values  # dN/dlogDp or dN

    # Auto-detect PSD type if needed
    if psd_type == 'auto':
        detected_type, confidence = _detect_psd_type(number_conc, diameter)
        psd_type = detected_type

        if confidence == 'low':
            warnings.warn(
                f"PSD type auto-detection has low confidence. "
                f"Assuming '{detected_type}'. Please specify psd_type explicitly "
                f"('dNdlogDp' or 'dN') to avoid incorrect results.",
                UserWarning
            )
        elif confidence == 'medium':
            warnings.warn(
                f"PSD type auto-detected as '{detected_type}' with medium confidence. "
                f"If results seem incorrect, try specifying psd_type explicitly.",
                UserWarning
            )
        # High confidence: no warning

    # Cross-sectional area × number concentration (scaled to Mm⁻¹)
    # π/4 * Dp² * N * 1e-6 (nm² to Mm⁻¹ conversion)
    cross_section_area = np.pi * (diameter / 2)**2 * number_conc * 1e-6

    # Get or calculate Mie efficiencies
    if precomputed_Q:
        Q_ext = precomputed_Q['Q_ext']
        Q_sca = precomputed_Q['Q_sca']
    else:
        _eff = calculate_mie_efficiencies(
            refractive_index, wavelength, diameter
        )
        Q_ext = _eff['Q_ext']
        Q_sca = _eff['Q_sca']

    # === Integrate over size distribution ===
    if multi_ri_per_psd:
        # Multiple refractive indices per PSD (for RI retrieval)
        n_times = len(psd)
        n_ri = len(refractive_index)

        # Expand arrays for broadcasting
        area_expanded = np.repeat(
            cross_section_area, n_ri, axis=0
        ).reshape(n_times, n_ri, -1)

        Q_ext_expanded = np.repeat(
            Q_ext[np.newaxis, :, :], n_times, axis=0
        ).reshape(n_times, n_ri, -1)

        Q_sca_expanded = np.repeat(
            Q_sca[np.newaxis, :, :], n_times, axis=0
        ).reshape(n_times, n_ri, -1)

        # Integrate based on psd_type
        integrand_ext = area_expanded * Q_ext_expanded
        integrand_sca = area_expanded * Q_sca_expanded

        if psd_type == 'dNdlogDp':
            log_dp = np.log10(diameter)
            ext_values = trapezoid(integrand_ext, x=log_dp, axis=-1)
            sca_values = trapezoid(integrand_sca, x=log_dp, axis=-1)
        else:  # dN
            ext_values = np.sum(integrand_ext, axis=-1)
            sca_values = np.sum(integrand_sca, axis=-1)

        extinction = pd.DataFrame(
            ext_values, columns=refractive_index, index=psd.index
        ).astype(float)

        scattering = pd.DataFrame(
            sca_values, columns=refractive_index, index=psd.index
        ).astype(float)

        absorption = extinction - scattering

        return {'ext': extinction, 'sca': scattering, 'abs': absorption}

    else:
        # Standard mode: one RI per time point
        integrand_ext = Q_ext * cross_section_area
        integrand_sca = Q_sca * cross_section_area

        result = pd.DataFrame(index=psd.index)
        result['ext'] = _integrate_psd(integrand_ext, diameter, psd_type).astype(float)
        result['sca'] = _integrate_psd(integrand_sca, diameter, psd_type).astype(float)
        result['abs'] = result['ext'] - result['sca']

        return result


# =============================================================================
# Additional Functions: Distribution, Mass Efficiency, Mixing Models
# =============================================================================

def calculate_extinction_distribution(
    refractive_index: complex | np.ndarray,
    wavelength: float,
    diameter: np.ndarray,
    number_conc: np.ndarray,
) -> dict[str, np.ndarray]:
    """
    Calculate extinction/scattering/absorption distribution per size bin.

    Unlike Mie_SD which integrates over all sizes, this function returns
    the contribution from each size bin (dExt/dlogDp).

    Parameters
    ----------
    refractive_index : complex or np.ndarray
        Complex refractive index. Can be:
        - Single complex value (applied to all)
        - Array of complex values (one per row of number_conc)
    wavelength : float
        Wavelength of incident light in nm.
    diameter : np.ndarray
        Particle diameters in nm. Shape: (n_bins,)
    number_conc : np.ndarray
        Number concentration (dN/dlogDp). Shape: (n_bins,) or (n_times, n_bins)

    Returns
    -------
    dict
        Dictionary with keys:
        - 'ext': Extinction distribution (dExt/dlogDp) in Mm⁻¹
        - 'sca': Scattering distribution (dSca/dlogDp) in Mm⁻¹
        - 'abs': Absorption distribution (dAbs/dlogDp) in Mm⁻¹
        - 'diameter': Particle diameters (nm)

    Examples
    --------
    >>> dp = np.logspace(1, 3, 50)
    >>> ndp = np.random.rand(50) * 1000
    >>> m = complex(1.5, 0.02)
    >>> dist = calculate_extinction_distribution(m, 550, dp, ndp)
    >>> print(dist['ext'].shape)  # (50,)

    Notes
    -----
    Output is in dExt/dlogDp units. To get total extinction:
        total_ext = np.trapz(dist['ext'], np.log10(diameter))
    """
    # Handle input dimensions
    number_conc = np.atleast_2d(number_conc)
    if number_conc.shape[1] != len(diameter):
        number_conc = number_conc.T

    n_times = number_conc.shape[0]

    # Handle refractive index
    if isinstance(refractive_index, complex):
        ri_array = np.array([refractive_index] * n_times)
    else:
        ri_array = np.atleast_1d(refractive_index)
        if len(ri_array) == 1:
            ri_array = np.array([ri_array[0]] * n_times)

    # Calculate Mie efficiencies
    _eff = calculate_mie_efficiencies(ri_array, wavelength, diameter)
    Q_ext, Q_sca = _eff['Q_ext'], _eff['Q_sca']

    # Cross-sectional area (π/4 * Dp²) in nm², scaled to Mm⁻¹
    cross_section = np.pi / 4 * diameter**2 * 1e-6

    # Calculate distributions: dX/dlogDp = Q * (π/4 * Dp²) * dN/dlogDp
    # Q_ext shape: (n_times, n_bins), cross_section shape: (n_bins,)
    # number_conc shape: (n_times, n_bins)
    ext_dist = Q_ext * cross_section * number_conc  # (n_times, n_bins)
    sca_dist = Q_sca * cross_section * number_conc

    abs_dist = ext_dist - sca_dist

    return {
        'ext': ext_dist.squeeze(),
        'sca': sca_dist.squeeze(),
        'abs': abs_dist.squeeze(),
        'diameter': diameter
    }


def calculate_mass_efficiency(
    refractive_index: complex,
    wavelength: float,
    diameter: np.ndarray,
    density: float
) -> dict[str, np.ndarray]:
    """
    Calculate mass extinction/scattering/absorption efficiency (MEE/MSE/MAE).

    Parameters
    ----------
    refractive_index : complex
        Complex refractive index (n + ik).
    wavelength : float
        Wavelength of incident light in nm.
    diameter : np.ndarray
        Particle diameters in nm.
    density : float
        Particle density in g/cm³.

    Returns
    -------
    dict
        Dictionary with keys:
        - 'MEE': Mass extinction efficiency (m²/g)
        - 'MSE': Mass scattering efficiency (m²/g)
        - 'MAE': Mass absorption efficiency (m²/g)
        - 'diameter': Particle diameters (nm)

    Examples
    --------
    >>> dp = np.logspace(1, 3, 50)
    >>> result = calculate_mass_efficiency(
    ...     complex(1.5, 0.02), wavelength=550, diameter=dp, density=1.5
    ... )
    >>> print(f"MEE at 100nm: {result['MEE'][25]:.2f} m²/g")

    Notes
    -----
    MEE = (3/2) * Q / (ρ * Dp) * 1000

    Where:
    - Q: Mie efficiency
    - ρ: particle density (g/cm³)
    - Dp: particle diameter (nm)
    - Factor 1000 converts to m²/g
    """
    # Calculate Q for single refractive index
    ri_array = np.array([refractive_index])
    _eff = calculate_mie_efficiencies(ri_array, wavelength, diameter)
    Q_ext, Q_sca = _eff['Q_ext'], _eff['Q_sca']
    # Q_ext shape: (1, n_bins), extract first row to get (n_bins,)

    # MEE = 3Q / (2ρDp) * 1000
    # Factor breakdown: 3/(2*ρ*Dp) where Dp in nm, ρ in g/cm³
    # Multiply by 1000 to get m²/g
    factor = 3 / (2 * density * diameter) * 1000

    MEE = Q_ext[0, :] * factor  # shape: (n_bins,)
    MSE = Q_sca[0, :] * factor
    MAE = MEE - MSE

    return {
        'MEE': MEE,
        'MSE': MSE,
        'MAE': MAE,
        'diameter': diameter
    }


# =============================================================================
# Mixing Models for Multi-Component Aerosols
# =============================================================================

# Default refractive indices for common aerosol species at 550 nm —
# sourced from core/_constants.py so this stays in sync with the
# Chemistry-side mass reconstruction values.
from AeroViz.dataProcess.core import REFRACTIVE_INDEX as _RI

DEFAULT_REFRACTIVE_INDICES = dict(_RI['550'])


def internal_mixing(
    psd: pd.DataFrame,
    refractive_index: pd.DataFrame | pd.Series,
    wavelength: float = 550,
    psd_type: str = 'auto',
) -> pd.DataFrame:
    """
    Calculate optical properties using internal mixing model.

    In internal mixing, all species are homogeneously mixed within each
    particle. The effective refractive index is the volume-weighted average.

    Parameters
    ----------
    psd : pd.DataFrame
        Particle size distribution.
        Columns: particle diameters (nm)
        Rows: time points
    refractive_index : pd.DataFrame or pd.Series
        Complex refractive index for each time point.
        Should have columns 'n' and 'k', or be complex values directly.
    wavelength : float, default=550
        Wavelength of incident light in nm.
    psd_type : str, default='auto'
        Type of PSD input:
        - 'dNdlogDp': Number concentration per log bin width (#/cm³)
        - 'dN': Number concentration per bin (#/cm³/bin)
        - 'auto': Auto-detect with warning if uncertain

    Returns
    -------
    pd.DataFrame
        Optical coefficients with columns: ext, sca, abs (Mm⁻¹)

    Examples
    --------
    >>> # PSD data
    >>> dp = np.logspace(1, 3, 50)
    >>> psd = pd.DataFrame(np.random.rand(10, 50) * 1000, columns=dp)
    >>>
    >>> # Refractive index (volume-weighted average)
    >>> ri = pd.DataFrame({'n': [1.52]*10, 'k': [0.01]*10})
    >>> result = internal_mixing(psd, ri, psd_type='dNdlogDp')
    """
    # Convert RI to complex array
    if isinstance(refractive_index, pd.DataFrame):
        if 'n' in refractive_index.columns and 'k' in refractive_index.columns:
            ri_array = (refractive_index['n'] + 1j * refractive_index['k']).values
        else:
            ri_array = refractive_index.iloc[:, 0].values
    else:
        ri_array = np.array(refractive_index)

    # Use standard Mie_SD calculation
    return Mie_SD(ri_array, wavelength, psd, psd_type=psd_type)


def external_mixing(
    psd: pd.DataFrame,
    volume_fractions: pd.DataFrame,
    wavelength: float = 550,
    refractive_indices: dict = None,
    psd_type: str = 'auto',
) -> pd.DataFrame:
    """
    Calculate optical properties using external mixing model.

    In external mixing, each species exists as separate particles.
    The total optical property is the sum of contributions from each species.

    Parameters
    ----------
    psd : pd.DataFrame
        Total particle size distribution.
        Columns: particle diameters (nm)
        Rows: time points
    volume_fractions : pd.DataFrame
        Volume fraction of each species. Columns should include:
        AS, AN, OM, Soil, SS, EC, (optional: ALWC)
    wavelength : float, default=550
        Wavelength of incident light in nm.
    refractive_indices : dict, optional
        Custom refractive indices for species. Default uses standard values.
    psd_type : str, default='auto'
        Type of PSD input:
        - 'dNdlogDp': Number concentration per log bin width (#/cm³)
        - 'dN': Number concentration per bin (#/cm³/bin)
        - 'auto': Auto-detect with warning if uncertain

    Returns
    -------
    pd.DataFrame
        Total optical coefficients with columns: ext, sca, abs (Mm⁻¹)

    Examples
    --------
    >>> dp = np.logspace(1, 3, 50)
    >>> psd = pd.DataFrame(np.random.rand(10, 50) * 1000, columns=dp)
    >>> vol_frac = pd.DataFrame({
    ...     'AS': [0.3]*10, 'AN': [0.2]*10, 'OM': [0.3]*10,
    ...     'Soil': [0.05]*10, 'SS': [0.05]*10, 'EC': [0.1]*10
    ... })
    >>> result = external_mixing(psd, vol_frac, psd_type='dNdlogDp')
    """
    if refractive_indices is None:
        refractive_indices = DEFAULT_REFRACTIVE_INDICES.copy()

    diameter = psd.columns.values.astype(float)
    n_times = len(psd)

    # Auto-detect PSD type if needed
    if psd_type == 'auto':
        detected_type, confidence = _detect_psd_type(psd.values, diameter)
        psd_type = detected_type

        if confidence == 'low':
            warnings.warn(
                f"PSD type auto-detection has low confidence. "
                f"Assuming '{detected_type}'. Please specify psd_type explicitly "
                f"('dNdlogDp' or 'dN') to avoid incorrect results.",
                UserWarning
            )
        elif confidence == 'medium':
            warnings.warn(
                f"PSD type auto-detected as '{detected_type}' with medium confidence. "
                f"If results seem incorrect, try specifying psd_type explicitly.",
                UserWarning
            )

    # Initialize result
    total_ext = np.zeros(n_times)
    total_sca = np.zeros(n_times)

    # Check for ALWC correction
    has_alwc = 'ALWC' in volume_fractions.columns
    if has_alwc:
        alwc_factor = 1 / (1 + volume_fractions['ALWC'].values)
    else:
        alwc_factor = np.ones(n_times)

    # Calculate contribution from each species
    for species, ri in refractive_indices.items():
        if species not in volume_fractions.columns:
            continue
        if species == 'ALWC':
            continue  # ALWC is handled separately

        vol_frac = volume_fractions[species].values

        # Species PSD = total PSD × volume fraction (with ALWC correction)
        species_psd = psd.values * (vol_frac * alwc_factor).reshape(-1, 1)

        # Calculate Mie for this species (single RI for all times)
        ri_array = np.array([ri] * n_times)
        _eff = calculate_mie_efficiencies(ri_array, wavelength, diameter)
        Q_ext, Q_sca = _eff['Q_ext'], _eff['Q_sca']
        # Q_ext shape: (n_times, n_bins)

        # Cross-sectional area
        cross_section = np.pi * (diameter / 2)**2 * 1e-6  # shape: (n_bins,)

        # Integrate over diameter bins
        # species_psd shape: (n_times, n_bins)
        integrand_ext = Q_ext * cross_section * species_psd  # (n_times, n_bins)
        integrand_sca = Q_sca * cross_section * species_psd
        total_ext += _integrate_psd(integrand_ext, diameter, psd_type)
        total_sca += _integrate_psd(integrand_sca, diameter, psd_type)

    # Build result DataFrame
    result = pd.DataFrame(index=psd.index)
    result['ext'] = total_ext.astype(float)
    result['sca'] = total_sca.astype(float)
    result['abs'] = (total_ext - total_sca).astype(float)

    return result


def generate_lognormal_psd(
    geometric_mean: float = 200,
    geometric_std: float = 2.0,
    total_number: float = 1e6,
    dp_range: tuple = (1, 2500),
    n_bins: int = 167,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Generate a lognormal particle size distribution.

    Parameters
    ----------
    geometric_mean : float, default=200
        Geometric mean diameter in nm.
    geometric_std : float, default=2.0
        Geometric standard deviation.
    total_number : float, default=1e6
        Total number concentration (#/cm³).
    dp_range : tuple, default=(1, 2500)
        Diameter range (min, max) in nm.
    n_bins : int, default=167
        Number of size bins.

    Returns
    -------
    diameter : np.ndarray
        Particle diameters in nm.
    ndp : np.ndarray
        Number concentration (dN/dlogDp).

    Examples
    --------
    >>> dp, ndp = generate_lognormal_psd(geometric_mean=100, geometric_std=1.8)
    >>> print(f"Peak at {dp[np.argmax(ndp)]:.1f} nm")
    """
    diameter = np.logspace(np.log10(dp_range[0]), np.log10(dp_range[1]), n_bins)

    # Lognormal distribution: dN/dlogDp
    log_sigma = np.log(geometric_std)
    log_dp = np.log(diameter)
    log_mean = np.log(geometric_mean)

    ndp = total_number * (
        1 / (log_sigma * np.sqrt(2 * np.pi)) *
        np.exp(-(log_dp - log_mean)**2 / (2 * log_sigma**2))
    )

    return diameter, ndp


# =============================================================================
# One-shot lognormal entry points (convenience over generate_lognormal_psd + Mie_SD)
# =============================================================================

def mie_lognormal(
    refractive_index,
    wavelength: float = 550,
    *,
    geo_mean: float = 200,
    geo_std: float = 2.0,
    total_number: float = 1e6,
    n_bins: int = 167,
    dp_range: tuple = (1, 2500),
) -> dict:
    """
    Compute Mie optics for a single lognormal PSD without building it manually.

    Equivalent to ``generate_lognormal_psd`` + ``Mie_SD`` in one call,
    returning a flat dict of scalars. For multi-modal distributions use
    :func:`mie_multimodal`.

    Parameters
    ----------
    refractive_index : complex
        Particle complex refractive index ``m = n + ik``.
    wavelength : float, default 550
        Wavelength in nm.
    geo_mean : float, default 200
        Geometric mean diameter in nm.
    geo_std : float, default 2.0
        Geometric standard deviation (dimensionless).
    total_number : float, default 1e6
        Total number concentration (#/cm³).
    n_bins : int, default 167
        Number of log-spaced bins.
    dp_range : tuple, default (1, 2500)
        Diameter range in nm.

    Returns
    -------
    dict
        Keys ``ext``, ``sca``, ``abs`` (Mm⁻¹) — bulk optical coefficients.
    """
    dp, ndp = generate_lognormal_psd(
        geometric_mean=geo_mean,
        geometric_std=geo_std,
        total_number=total_number,
        dp_range=dp_range,
        n_bins=n_bins,
    )
    psd = pd.DataFrame([ndp], columns=dp)
    ri = np.atleast_1d(np.asarray(refractive_index, dtype=complex))
    result_df = Mie_SD(ri, wavelength, psd, psd_type='dNdlogDp')
    return result_df.iloc[0].to_dict()


def mie_multimodal(
    refractive_index,
    wavelength: float = 550,
    *,
    modes,
    n_bins: int = 167,
    dp_range: tuple = (1, 2500),
) -> dict:
    """
    Compute Mie optics for a multi-modal lognormal PSD.

    Common atmospheric setup: nucleation + Aitken + accumulation +
    coarse modes summed on a single diameter grid.

    Parameters
    ----------
    refractive_index : complex
        Particle complex refractive index ``m = n + ik``.
    wavelength : float, default 550
        Wavelength in nm.
    modes : sequence of (geo_mean, geo_std, total_number)
        Each entry defines one lognormal mode (nm, dimensionless, #/cm³).
        Example: ``[(50, 1.5, 1e4), (200, 2.0, 5e3)]`` for an Aitken +
        accumulation mode.
    n_bins : int, default 167
        Number of log-spaced bins for the merged grid.
    dp_range : tuple, default (1, 2500)
        Diameter range in nm covering all modes.

    Returns
    -------
    dict
        Keys ``ext``, ``sca``, ``abs`` (Mm⁻¹).
    """
    if not modes:
        raise ValueError("mie_multimodal() needs at least one mode in `modes`.")

    diameter = np.logspace(np.log10(dp_range[0]), np.log10(dp_range[1]), n_bins)
    log_dp = np.log(diameter)
    total_ndp = np.zeros(n_bins)

    for gm, gs, nt in modes:
        log_sigma = np.log(gs)
        log_mean = np.log(gm)
        ndp = nt * (
            1 / (log_sigma * np.sqrt(2 * np.pi))
            * np.exp(-(log_dp - log_mean) ** 2 / (2 * log_sigma ** 2))
        )
        total_ndp += ndp

    psd = pd.DataFrame([total_ndp], columns=diameter)
    ri = np.atleast_1d(np.asarray(refractive_index, dtype=complex))
    result_df = Mie_SD(ri, wavelength, psd, psd_type='dNdlogDp')
    return result_df.iloc[0].to_dict()


# =============================================================================
# Backward Compatibility Aliases
# =============================================================================

MieQ = calculate_mie_efficiencies   # modern alias — returns dict


def Mie_Q(m, wavelength, dp):
    """PyMieScatt-compatible 7-tuple alias for
    :func:`calculate_mie_efficiencies`.

    Mirrors the legacy ``mie_theory.Mie_Q`` (and upstream
    ``PyMieScatt.MieQ``) signature, returning::

        (Q_ext, Q_sca, Q_abs, g, Q_pr, Q_back, Q_ratio)

    New code should prefer :func:`calculate_mie_efficiencies` (dict).
    """
    d = calculate_mie_efficiencies(m, wavelength, dp)
    return (d['Q_ext'], d['Q_sca'], d['Q_abs'],
            d['g'], d['Q_pr'], d['Q_back'], d['Q_ratio'])


Mie_ab = calculate_mie_coefficients


def Mie_PESD(m, wavelength=550, dp=None, ndp=None, **kwargs):
    """Backward-compatible 3-tuple wrapper for
    :func:`calculate_extinction_distribution`.

    Mirrors the legacy ``mie_theory.Mie_PESD`` 3-tuple interface
    ``(ext, sca, abs)`` so existing callers can swap the import path
    without rewriting their unpacking. New code should prefer
    :func:`calculate_extinction_distribution` (returns a dict).
    """
    result = calculate_extinction_distribution(m, wavelength, dp, ndp, **kwargs)
    return result['ext'], result['sca'], result['abs']


def Mie_MEE(m, wavelength, dp, density):
    """Backward-compatible 3-tuple wrapper for
    :func:`calculate_mass_efficiency`.

    Mirrors the legacy ``mie_theory.Mie_MEE`` 3-tuple interface
    ``(MEE, MSE, MAE)``. New code should prefer
    :func:`calculate_mass_efficiency` (returns a dict).
    """
    result = calculate_mass_efficiency(m, wavelength, dp, density)
    return result['MEE'], result['MSE'], result['MAE']



"""Tests for Q-ACSM (Aerosol Chemical Speciation Monitor) reader."""
import pytest
from .base import BaseReaderTest


@pytest.mark.q_acsm
class TestQACSMReader(BaseReaderTest):
    INSTRUMENT = 'Q-ACSM'
